from abc import ABC, abstractmethod


class IDaprClientAdapter(ABC):
    """ Interface for adapting the Dapr client operations. """

    @abstractmethod
    def start_workflow(self, workflow_component, workflow_name, input):
        pass

    @abstractmethod
    def get_workflow(self, instance_id, workflow_component):
        pass

    @abstractmethod
    def register_workflow(self, workflow):
        pass

    @abstractmethod
    def register_activity(self, activity):
        pass

    @abstractmethod
    def raise_workflow_event(self, instance_id, workflow_component, event_name, event_data):
        pass



"""

"""
import dspy
from dspygen.utils.dspy_tools import init_dspy        


class LongFormQAModule(dspy.Module):
    """LongFormQAModule"""
    
    def __init__(self, **forward_args):
        super().__init__()
        self.forward_args = forward_args
        self.output = None
        
    def __or__(self, other):
        if other.output is None and self.output is None:
            self.forward(**self.forward_args)

        other.pipe(self.output)

        return other

    def forward(self, context, question):
        pred = dspy.Predict("context, question -> query")
        self.output = pred(context=context, question=question).query
        return self.output
        
    def pipe(self, input_str):
        raise NotImplementedError("Please implement the pipe method for DSL support.")
        # Replace TODO with a keyword from you forward method
        # return self.forward(TODO=input_str)


from typer import Typer
app = Typer()


@app.command()
def call(context, question):
    """LongFormQAModule"""
    init_dspy()

    print(long_form_qa_call(context=context, question=question))



def long_form_qa_call(context, question):
    long_form_qa = LongFormQAModule()
    return long_form_qa.forward(context=context, question=question)



def main():
    init_dspy()
    context = ""
    question = ""
    print(long_form_qa_call(context=context, question=question))



from fastapi import APIRouter
router = APIRouter()

@router.post("/long_form_qa/")
async def long_form_qa_route(data: dict):
    # Your code generation logic here
    init_dspy()

    print(data)
    return long_form_qa_call(**data)



"""
import streamlit as st


# Streamlit form and display
st.title("LongFormQAModule Generator")
context = st.text_input("Enter context")
question = st.text_input("Enter question")

if st.button("Submit LongFormQAModule"):
    init_dspy()

    result = long_form_qa_call(context=context, question=question)
    st.write(result)
"""

if __name__ == "__main__":
    main()



"""

"""
import dspy
from dspygen.utils.dspy_tools import init_dspy


class BusinessRequirementsSignature(dspy.Signature):
    """
    Transforms business requirement descriptions provided in BPMN format into DMN YAML format.
    """

    bpmn = dspy.InputField(desc="Business Process Model and Notation as input")

    dmn_yaml = dspy.OutputField(desc="Decision Model and Notation in YAML format as output", prefix="```yaml")


class BusinessRequirementsModule(dspy.Module):
    """BusinessRequirementsModule"""

    def forward(self, bpmn):
        pred = dspy.ChainOfThought(BusinessRequirementsSignature)
        result = pred(bpmn=bpmn).dmn_yaml
        return result


from typer import Typer
app = Typer()


@app.command()
def call(bpmn):
    """BusinessRequirementsModule"""
    init_dspy()

    print(business_requirements_call(bpmn=bpmn))



def business_requirements_call(bpmn):
    business_requirements = BusinessRequirementsModule()
    return business_requirements.forward(bpmn=bpmn)


bpmn = """The project must integrated the shippiing labels produced by USP ConnectShip shipping station with the certification number generated by the decision tree questionnaire.

At the time of the shipping lable being produced, the event should halt moving to the next screen of the lable pritning process and invoke the browser.  The browser loads order specific questionnariere described as a decision tree and allow technician t provide answers about the order being packaed.

Once the questionnaire is complete, the browser s closed, the questionnare cerificate id be posted and included to be printed to shipping label.

The decision tree selection is based on types of products in the order and should be managed / created by business analysts."""


def main():
    init_dspy()
    print(business_requirements_call(bpmn=bpmn))



from fastapi import APIRouter
router = APIRouter()

@router.post("/business_requirements/")
async def business_requirements_route(data: dict):
    # Your code generation logic here
    init_dspy()

    print(data)
    return business_requirements_call(**data)



"""
import streamlit as st


# Streamlit form and display
st.title("BusinessRequirementsModule Generator")
bpmn = st.text_input("Enter bpmn")

if st.button("Submit BusinessRequirementsModule"):
    init_dspy()

    result = business_requirements_call(bpmn=bpmn)
    st.write(result)
"""

if __name__ == "__main__":
    main()



# conversations.py
from abc import ABC, abstractmethod


class QuestionHandler(ABC):
    @abstractmethod
    def parse_question(self, question: str) -> dict:
        pass


class SocraticQuestionHandler(QuestionHandler):
    def parse_question(self, question: str) -> dict:
        """Parse a question and return the interpreted data."""


# Implement Socratic question parsing here

class ProjectBasedLearningHandler(QuestionHandler):
    def parse_question(self, question: str) -> dict:
        """Parse a question and return the interpreted data."""


# Implement project-based learning question parsing here

class ResponseGenerator(ABC):
    @abstractmethod
    def generate_response(self, interpreted_data: dict, user_data: dict) -> str:
        pass


class ContextualResponseGenerator(ResponseGenerator):
    def generate_response(self, interpreted_data: dict, user_data: dict) -> str:
        """Generate a response based on the interpreted data and user data."""


# Implement contextual response generation here

class LearningSession:
    def __init__(self, question_handler: QuestionHandler, response_generator: ResponseGenerator):
        self.question_handler = question_handler
        self.response_generator = response_generator

    def start(self, question: str):
        interpreted_data = self.question_handler.parse_question(question)
        # response = self.response_generator.generate_response(interpreted_data, user_data)
        # Add interactivity by returning or presenting the response


# app.py


def start_conversation(project_dependencies=None, conversations=None) -> None:
    question_handler = project_dependencies.create_question_handler()
    response_generator = project_dependencies.create_response_generator()

    learning_session = conversations.LearningSession(question_handler, response_generator)
    while True:
        user_question = input("Ask a question: ")
        learning_session.start(user_question)


if __name__ == "__main__":
    start_conversation()


import ijson


from typing import Optional

from pydantic import BaseModel, ValidationError


class Author(BaseModel):
    role: str
    name: Optional[str] = None
    metadata: dict


class ContentPart(BaseModel):
    content_type: str
    parts: list[str] | None


class Message(BaseModel):
    id: str
    author: Author
    content: ContentPart
    status: str
    metadata: dict


class Data(BaseModel):
    id: str
    message: Message | None
    parent: str | None
    children: list[str]


class Conversation(BaseModel):
    title: str
    mapping: dict


# Function to process each conversation chunk
def process_conversations_chunk(chunk):
    # Define Pydantic models as before

    # Process each conversation in the chunk
    for chunked in chunk:
        try:
            conversation = Conversation(**chunked)
            # Do whatever processing you need with the conversation
            # print(f"Title: {conversation.title}")
            for key in conversation.mapping:
                data = Data(**conversation.mapping[key])
                if data.message and data.message.author.role == "assistant":
                    for part in data.message.content.parts:
                        if "Technology Applications Inc" in part:
                            print(part)
                        # encoding = tiktoken.encoding_for_model("text-embedding-ada-002")
                        # print(len(encoding.encode(part)))
                        # print(part)
        except ValidationError:
            # print(e)
            pass


def main():
    # Define the path to your large JSON file
    json_file_path = "/Users/sac/dev/dspygen/data/conversations.json"

    # Open the JSON file for streaming
    with open(json_file_path, "rb") as json_file:
        conversations_generator = ijson.items(
            json_file, "item"
        )  # Assumes each conversation is a separate JSON object

        # Process the conversations in chunks (adjust the chunk size as needed)
        chunk_size = 10  # Define your desired chunk size
        chunk = []
        for conversation in conversations_generator:
            chunk.append(conversation)
            if len(chunk) >= chunk_size:
                process_conversations_chunk(chunk)
                chunk = []

        # Process any remaining conversations in the last chunk
        if chunk:
            process_conversations_chunk(chunk)


if __name__ == "__main__":
    main()



"""

"""
import dspy
from dspygen.utils.dspy_tools import init_dspy


class FunctionInvokeModule(dspy.Module):
    """FunctionInvokeModule"""

    def __init__(self, **forward_args):
        super().__init__()
        self.forward_args = forward_args
        self.output = None

    def __or__(self, other):
        if other.output is None and self.output is None:
            self.forward(**self.forward_args)

        other.pipe(self.output)

        return other

    def forward(self, function_declaration):
        pred = dspy.Predict("function_declaration -> invocation_command")
        self.output = pred(function_declaration=function_declaration).invocation_command
        return self.output

    def pipe(self, input_str):
        raise NotImplementedError("Please implement the pipe method for DSL support.")
        # Replace TODO with a keyword from you forward method
        # return self.forward(TODO=input_str)


from typer import Typer

app = Typer()


@app.command()
def call(function_declaration):
    """FunctionInvokeModule"""
    init_dspy()

    print(function_invoke_call(function_declaration=function_declaration))


def function_invoke_call(function_declaration):
    function_invoke = FunctionInvokeModule()
    return function_invoke.forward(function_declaration=function_declaration)


quick = """def quicksort(arr):
    if len(arr) <= 1:
        return arr
    pivot = arr[0]
    left = [x for x in arr[1:] if x < pivot]
    right = [x for x in arr[1:] if x >= pivot]
    return quicksort(left) + [pivot] + quicksort(right)"""


def main():
    init_dspy()
    function_declaration = quick
    result = function_invoke_call(function_declaration=function_declaration)
    print(result)
    exec(function_declaration)
    eval(result)


from fastapi import APIRouter

router = APIRouter()


@router.post("/function_invoke/")
async def function_invoke_route(data: dict):
    # Your code generation logic here
    init_dspy()

    print(data)
    return function_invoke_call(**data)


"""
import streamlit as st


# Streamlit form and display
st.title("FunctionInvokeModule Generator")
function_declaration = st.text_input("Enter function_declaration")

if st.button("Submit FunctionInvokeModule"):
    init_dspy()

    result = function_invoke_call(function_declaration=function_declaration)
    st.write(result)
"""

if __name__ == "__main__":
    main()



"""The BaseActor Module for Reactive Domain-Driven Design (RDDDY) Framework
---------------------------------------------------------------------

This module implements the core Actor abstraction within the RDDDY framework, providing a robust foundation for building reactive, domain-driven systems that are scalable, maintainable, and capable of handling complex, concurrent interactions. The Actor model encapsulates both state and behavior, allowing for asynchronous message passing as the primary means of communication between actors, thus fostering loose coupling and enhanced system resilience.

### Overview

Actors are the fundamental units of computation in the RDDDY framework. Each actor possesses a unique identity, a mailbox for message queuing, and a set of behaviors to handle incoming messages. The Actor module facilitates the creation, supervision, and coordination of actors within an ActorSystem, ensuring that messages are delivered and processed in a manner consistent with the system's domain-driven design principles.

### Multiline Calculus for Actor Behavior Specification

The operational semantics of actors within the RDDDY framework are formalized through a rigorous multiline calculus, which outlines the preconditions, postconditions, and invariants that govern actor behavior and interaction. This calculus serves as a contract, ensuring that actors behave correctly and predictably within their designated domain contexts.

#### Actor State Transition

Given an actor \(A\) with state \(s\) and a message \(m\), the state transition is defined as:

Precondition ( Pre Pre): 𝑠 ∈ 𝑆 s∈S and 𝑚 ∈ 𝑀 m∈M
Transition: 𝑇 ( 𝑠 , 𝑚 ) → 𝑠 ′ T(s,m)→s ′
Postcondition ( Post Post): 𝑠 ′ ∈ 𝑆 ′ s ′ ∈S ′

#### Message Handling

For a message \(m\) handled by the actor, leading to a state modification:

Precondition ( Pre Pre): 𝑠 ∈ 𝑆 s∈S and 𝑚 ∈ 𝑀 m∈M
Handling: 𝐻 ( 𝑚 , 𝑠 ) → 𝑠 ′ ′ H(m,s)→s ′′
Postcondition ( Post Post): 𝑠 ′ ′ ∈ 𝑆 ′ ′ s ′′ ∈S ′′

#### Invariant Maintenance

Ensuring system invariants \(I\) across transitions:

Invariant: 𝐼 ( 𝑆 ) ∧ 𝑇 ( 𝑠 , 𝑚 ) → 𝑠 ′ ⇒ 𝐼 ( 𝑆 ′ ) I(S)∧T(s,m)→s ′ ⇒I(S ′ )

#### Domain-Specific Assertions

Linking actor state transitions to domain logic:

Precondition ( Pre Pre): Δ ( 𝑠 ) Δ(s) and 𝛿 ( 𝑚 ) δ(m)
Domain Logic: 𝐷 ( 𝑠 , 𝑚 ) → Δ ( 𝑠 ′ ) D(s,m)→Δ(s ′ )
Postcondition ( Post Post): Δ ′ ( 𝑠 ′ ) Δ ′ (s ′ )

### Purpose

This calculus not only specifies the expected behavior of actors in response to messages but also integrates domain-specific knowledge, ensuring that actors operate in alignment with the broader domain-driven objectives of the system. By adhering to these specifications, the Actor module provides a reliable and expressive framework for developing systems that are both technically sound and closely aligned with domain requirements.

### Usage

Developers should implement actor behaviors in accordance with the outlined calculus, ensuring that each actor's implementation respects the preconditions, postconditions, and domain-specific assertions relevant to their system's domain logic. This approach facilitates the development of systems that are not only functionally correct but also domain-compliant, thereby enhancing the value and applicability of the RDDDY framework in real-world scenarios.
"""
import asyncio
from collections.abc import Callable
from typing import TYPE_CHECKING, Optional, Type

import reactivex as rx
from loguru import logger
from reactivex import operators as ops
from reactivex.scheduler.eventloop import AsyncIOScheduler

from dspygen.rdddy.base_event import BaseEvent
from dspygen.rdddy.base_message import *

if TYPE_CHECKING:
    from dspygen.rdddy.actor_system import ActorSystem


class BaseActor:
    """Represents an actor within the RDDDY framework.

    Actors are fundamental units of computation in the RDDDY framework, encapsulating both state and behavior.
    They communicate asynchronously through message passing, promoting loose coupling and system resilience.

    Args:
        actor_system (ActorSystem): The ActorSystem to which the actor belongs.
        actor_id (int, optional): The unique identifier of the actor. Defaults to None.

    Attributes:
        actor_system (ActorSystem): The ActorSystem to which the actor belongs.
        actor_id (int): The unique identifier of the actor.
        mailbox (Subject): A subject for message queuing.
        handlers (dict): A mapping of message types to corresponding handler methods.

    Methods:
        start(scheduler): Starts the actor's mailbox processing loop.
        on_next(message): Callback function for processing incoming messages.
        on_error(error): Callback function for handling errors in the actor's mailbox.
        on_completed(): Callback function when the actor's mailbox stream completes.
        receive(message): Processes an incoming message.
        send(recipient_id, message): Sends a message to another actor.
        publish(message): Publishes a message to the actor system.
        map_handlers(): Maps message types to corresponding handler methods.
    """

    def __init__(self, actor_system: "ActorSystem", actor_id: Optional[int] = None):
        self.actor_system = actor_system
        self.actor_id = actor_id or id(self)
        self.mailbox = rx.subject.Subject()
        self.handlers = self.map_handlers()

    async def start(self, scheduler: AsyncIOScheduler):
        """Initiates the processing loop for the actor's mailbox, ensuring asynchronous message handling.

        Preconditions (Pre):
            - The actor's mailbox must be initialized.
            - A valid scheduler must be provided.

        Transition (T):
            - Initiates the processing loop for the actor's mailbox, enabling asynchronous message handling.

        Postconditions (Post):
            - The actor's mailbox processing loop has started successfully.

        Args:
            scheduler: An asynchronous scheduler used to control the execution of tasks.

        """
        self.mailbox.pipe(ops.observe_on(scheduler)).subscribe(
            on_next=self.on_next,  # Synchronous wrapper for async handler
            on_error=self.on_error,
            on_completed=self.on_completed,
        )
        logger.info(f"Actor {self.actor_id} started")

    def on_next(self, message: BaseMessage):
        """Handles the next incoming message in the actor's mailbox.

        Preconditions (Pre):
            - The incoming message must be a valid instance of the Message class.

        Transition (T):
            - Processes the incoming message asynchronously.

        Postconditions (Post):
            - The incoming message has been processed by the actor.

        Args:
            message (BaseMessage): The incoming message to be processed.
        """
        # Schedule the async handler as a new task
        # logger.debug(f"Actor {self.actor_id} received message: {message}")
        asyncio.create_task(self.receive(message))

    def on_error(self, error):
        """Handles errors that occur in the actor's mailbox processing.

        Preconditions (Pre):
            - None

        Transition (T):
            - Handles the error generated during mailbox processing.

        Postconditions (Post):
            - The error has been handled, and appropriate action has been taken.

        Args:
            error: The error object representing the error that occurred.
        """
        logger.error(f"Error in actor {self.actor_id} mailbox: {error}")

    def on_completed(self):
        """Handles the completion of the actor's mailbox stream.

        Preconditions (Pre):
            - None

        Transition (T):
            - Handles the completion event of the actor's mailbox stream.

        Postconditions (Post):
            - The actor's mailbox stream has completed, and appropriate action has been taken.
        """
        # logger.debug(f"Actor {self.actor_id} mailbox stream completed")

    async def receive(self, message: BaseMessage):
        """Processes an incoming message received by the actor.

        Preconditions (Pre):
            - The incoming message must be a valid instance of the Message class.

        Transition (T):
            - Processes the incoming message asynchronously, invoking the appropriate handler method.

        Postconditions (Post):
            - The incoming message has been successfully processed by the actor.

        Args:
            message (BaseMessage): The incoming message to be processed.
        """
        try:
            handler = self.handlers.get(type(message))
            if handler:
                logger.debug(
                    f"Actor handling message: {message} with {handler.__name__}"
                )
                await handler(message)
        except Exception as e:
            error_message = f"Error in actor {self.actor_id} processing message: {e}"
            # Broadcast an error event through the actor system
            await self.publish(BaseEvent(content=error_message))
            logger.error(error_message)

    async def publish(self, message: BaseMessage):
        """Publishes a message to the actor system for distribution.

        Preconditions (Pre):
            - The message must be a valid instance of the Message class.

        Transition (T):
            - Publishes the message to the actor system for distribution.

        Postconditions (Post):
            - The message has been successfully published to the actor system.

        Args:
            message (BaseMessage): The message to be published.
        """
        if message.actor_id == -1:
            message.actor_id = self.actor_id

        await self.actor_system.publish(message)

    def map_handlers(self) -> dict[Type[BaseMessage], Callable]:
        """Maps message types to corresponding handler methods.

        Preconditions (Pre):
            - None

        Transition (T):
            - Iterates through the methods of the actor instance and identifies callable methods with annotations.
            - Maps message types to corresponding handler methods based on method annotations.

        Postconditions (Post):
            - A dictionary containing message types as keys and corresponding handler methods as values has been generated.
        """
        handlers = {}
        for name, method in inspect.getmembers(self):
            if callable(method) and hasattr(method, "__annotations__"):
                annotations = method.__annotations__
                for arg in annotations.values():
                    try:
                        if issubclass(arg, BaseMessage):
                            handlers[arg] = method
                    except TypeError:
                        pass
        del handlers[BaseMessage]
        return handlers



"""

"""
import dspy
from dspygen.utils.dspy_tools import init_dspy        


class BinaryOutputModule(dspy.Module):
    """BinaryOutputModule"""

    def forward(self, requirements):
        pred = dspy.Predict("requirements -> binary_code")
        result = pred(requirements=requirements).binary_code
        return result


from typer import Typer
app = Typer()


@app.command()
def call(requirements):
    """BinaryOutputModule"""
    init_dspy()

    print(binary_output_call(requirements=requirements))



def binary_output_call(requirements):
    binary_output = BinaryOutputModule()
    return binary_output.forward(requirements=requirements)



def main():
    init_dspy()
    requirements = "C header file example"
    print(binary_output_call(requirements=requirements))



from fastapi import APIRouter
router = APIRouter()

@router.post("/binary_output/")
async def binary_output_route(data: dict):
    # Your code generation logic here
    init_dspy()

    print(data)
    return binary_output_call(**data)



"""
import streamlit as st


# Streamlit form and display
st.title("BinaryOutputModule Generator")
requirements = st.text_input("Enter requirements")

if st.button("Submit BinaryOutputModule"):
    init_dspy()

    result = binary_output_call(requirements=requirements)
    st.write(result)
"""

if __name__ == "__main__":
    main()



"""

"""
import dspy
from dspygen.utils.dspy_tools import init_dspy        


class CobolToPythonModule(dspy.Module):
    """CobolToPythonModule"""

    def forward(self, cobol):
        pred = dspy.Predict("cobol -> python")
        result = pred(cobol=cobol).python
        return result


from typer import Typer
app = Typer()


@app.command()
def call(cobol):
    """CobolToPythonModule"""
    init_dspy()

    print(cobol_to_python_call(cobol=cobol))



def cobol_to_python_call(cobol):
    cobol_to_python = CobolToPythonModule()
    return cobol_to_python.forward(cobol=cobol)



def main():
    init_dspy()
    cobol = ""
    print(cobol_to_python_call(cobol=cobol))



from fastapi import APIRouter
router = APIRouter()

@router.post("/cobol_to_python/")
async def cobol_to_python_route(data: dict):
    # Your code generation logic here
    init_dspy()

    print(data)
    return cobol_to_python_call(**data)



"""
import streamlit as st


# Streamlit form and display
st.title("CobolToPythonModule Generator")
cobol = st.text_input("Enter cobol")

if st.button("Submit CobolToPythonModule"):
    init_dspy()

    result = cobol_to_python_call(cobol=cobol)
    st.write(result)
"""

if __name__ == "__main__":
    main()



import re

import dspy
from dspy.evaluate import Evaluate
from dspy.datasets import HotPotQA
from dsp.utils import EM
from dspy.teleprompt import BayesianSignatureOptimizer, MIPRO

from dspygen.utils.dspy_tools import init_ol


colbert_v2_endpoint = "http://20.102.90.50:2017/wiki17_abstracts"
colbertv2 = dspy.ColBERTv2(url=colbert_v2_endpoint)


lm = init_ol()

dspy.settings.configure(rm=colbertv2, lm=lm)


class ReturnRankedDocuments(dspy.Signature):
    """Given a question we are trying to answer and a list of passages, return a comma separated list of the numbers associated with each passage. These numbers should be ordered by helpfulness in answering the question, with most helpful passage number first, and the least helpful last."""
    question = dspy.InputField(desc="The question we're trying to answer.")
    context = dspy.InputField(desc="List of potentially related passages.")
    ranking = dspy.OutputField(
        desc="A comma separated list of numbers corresponding to passage indices, ranked in descending order by their helpfulness in answering our question.")


class RankingMultiHop(dspy.Module):
    def __init__(self, hops, num_passages_to_retrieve, max_passages_in_context):
        super().__init__()
        self.hops = hops
        self.num_passages_to_retrieve = num_passages_to_retrieve
        self.max_passages_in_context = max_passages_in_context
        self.retrieve = dspy.Retrieve(k=self.num_passages_to_retrieve)
        self.generate_query = dspy.ChainOfThought("context ,question->search_query")
        self.generate_answer = dspy.ChainOfThought("context ,question->answer")
        self.generate_ranking = dspy.ChainOfThought(ReturnRankedDocuments)

    def forward(self, question):
        context = []
        full_context = []
        top_context = []
        max_passage_num = self.max_passages_in_context
        for hop in range(self.hops):
            # Get a new query
            query = self.generate_query(context=context, question=question).search_query
            # Get new passages
            context = self.retrieve(query).passages
            # Add these new passages to the previous top context
            full_context = top_context + context
            # Get the most important indices, ranked
            most_important_indices = self.generate_ranking(question=question, context=full_context).ranking
            indices = [int(num) for num in re.findall(r'\d+', most_important_indices)]

            if len(indices) < max_passage_num:
                indices = range(1, max_passage_num + 1)

            valid_indices = [index - 1 for index in indices if index - 1 < len(context)]
            top_indices = sorted(valid_indices, key=lambda x: x)[:max_passage_num + 1]
            most_important_context_list = [context[idx] for idx in top_indices]
            # Save the top context
            top_context = most_important_context_list

        return dspy.Prediction(context=context,
                               answer=self.generate_answer(context=top_context, question=question).answer)


def main():
    """Main function"""
    program = RankingMultiHop(hops=4, num_passages_to_retrieve=5, max_passages_in_context=5)

    # Load and configure the datasets.
    TRAIN_SIZE = 5
    EVAL_SIZE = 5

    hotpot_dataset = HotPotQA(train_seed=1, eval_seed=2023, test_size=0)
    trainset = [x.with_inputs('question') for x in hotpot_dataset.train][:TRAIN_SIZE]
    devset = [x.with_inputs('question') for x in hotpot_dataset.dev][:EVAL_SIZE]

    # Set up metrics
    NUM_THREADS = 10

    metric = dspy.evaluate.answer_exact_match

    kwargs = dict(num_threads=NUM_THREADS, display_progress=True)
    evaluate = Evaluate(devset=devset, metric=metric, **kwargs)

    # baseline_train_score = evaluate(program, devset=trainset)
    # baseline_eval_score = evaluate(program, devset=devset)

    # Define hyperparameters:
    N = 10  # The number of instructions and fewshot examples that we will generate and optimize over
    trials = 30  # The number of optimization trials to be run (we will test out a new combination of instructions and fewshot examples in each trial)
    temperature = 1.0  # The temperature configured for generating new instructions

    # Compile
    eval_kwargs = dict(num_threads=16, display_progress=True, display_table=0)
    teleprompter = MIPRO(prompt_model=lm, task_model=lm, metric=metric, num_candidates=N,
                         init_temperature=temperature, verbose=True)
    compiled_program = teleprompter.compile(program, trainset=trainset, num_trials=trials, max_bootstrapped_demos=1,
                                            max_labeled_demos=2, eval_kwargs=eval_kwargs)

    best_score = 0

    def get_signature(predictor):
        if (hasattr(predictor, 'extended_signature')):
            return predictor.extended_signature
        elif (hasattr(predictor, 'signature')):
            return predictor.signature

    print(f"Basline program | Score: {best_score}:")
    for i, predictor in enumerate(program.predictors()):
        print(f"Prompt {i + 1} Instruction: {get_signature(predictor).instructions}")
    print()

    print("----------------")

    for trial_num in compiled_program.trial_logs:
        program_score = compiled_program.trial_logs[trial_num]["score"]
        program_pruned = compiled_program.trial_logs[trial_num]["pruned"]
        if program_score > best_score and not program_pruned:
            best_score = program_score
            best_program_so_far = compiled_program.trial_logs[trial_num]["program"]
        if trial_num % 5 == 0:
            print(f"Best program after {trial_num} trials | Score: {best_score}:")
            for i, predictor in enumerate(best_program_so_far.predictors()):
                print(f"Prompt {i + 1} Instruction: {get_signature(predictor).instructions}")
            print()


if __name__ == '__main__':
    main()



import asyncio
from transitions import Machine
from transitions.extensions.states import add_state_features, Tags
from enum import Enum, auto

from dspygen.rdddy.base_actor import BaseActor
from dspygen.rdddy.base_command import BaseCommand
from dspygen.rdddy.base_event import BaseEvent
from dspygen.rdddy.actor_system import ActorSystem
from dspygen.utils.dspy_tools import init_dspy


# Enhancing state features for more complex state behavior
@add_state_features(Tags)
class CustomStateMachine(Machine):
    pass


# Enum for Dialogue States
class SocratesState(Enum):
    IDLE = auto()
    ASKING_QUESTION = auto()
    EVALUATING_RESPONSE = auto()
    CHALLENGING = auto()
    FOLLOW_UP = auto()
    CONCLUDING = auto()


class StudentState(Enum):
    IDLE = auto()
    RESPONDING = auto()
    CONSIDERING_FOLLOW_UP = auto()
    REVIEWING_FEEDBACK = auto()


# Commands
class AskQuestionCommand(BaseCommand):
    """Socrates asks a question."""


class AnswerCommand(BaseCommand):
    """Student provides an answer."""


class EvaluateResponseCommand(BaseCommand):
    """Evaluate the student's response."""


class ChallengeCommand(BaseCommand):
    """Issue a challenge based on the response."""


class FollowUpQuestionCommand(BaseCommand):
    """Ask a follow-up question."""


class ConcludeDialogueCommand(BaseCommand):
    """Conclude the dialogue positively."""


# Events
class QuestionAskedEvent(BaseEvent):
    """Question has been asked."""


class AnswerReceivedEvent(BaseEvent):
    """Answer has been received from the student."""


class ResponseEvaluatedEvent(BaseEvent):
    """Response has been evaluated."""


class ChallengeIssuedEvent(BaseEvent):
    """Challenge has been issued to the student."""


class FollowUpAskedEvent(BaseEvent):
    """Follow-up question has been asked."""


class ConcludeDialogueEvent(BaseEvent):
    """Dialogue has been concluded."""


class SocratesActor(BaseActor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.machine = CustomStateMachine(model=self, states=SocratesState, initial=SocratesState.IDLE)
        self.machine.add_transition(trigger='ask_question', source=SocratesState.IDLE, dest=SocratesState.ASKING_QUESTION)
        self.machine.add_transition(trigger='evaluate_response', source=SocratesState.ASKING_QUESTION, dest=SocratesState.EVALUATING_RESPONSE)
        self.machine.add_transition(trigger='issue_challenge', source=SocratesState.EVALUATING_RESPONSE, dest=SocratesState.CHALLENGING)
        self.machine.add_transition(trigger='ask_follow_up', source=SocratesState.CHALLENGING, dest=SocratesState.FOLLOW_UP)
        self.machine.add_transition(trigger='conclude_dialogue', source=SocratesState.FOLLOW_UP, dest=SocratesState.CONCLUDING)

    async def handle_AskQuestionCommand(self, message: AskQuestionCommand):
        question = "What led you to this conclusion about unchanged prices?"
        await self.publish(QuestionAskedEvent(content=question))

    async def handle_EvaluateResponseCommand(self, message: EvaluateResponseCommand):
        evaluation = "Evaluating the student's response..."
        await self.publish(ResponseEvaluatedEvent(content=evaluation))

    async def handle_ChallengeCommand(self, message: ChallengeCommand):
        challenge = "How do you reconcile your view with data on local climate effects?"
        await self.publish(ChallengeIssuedEvent(content=challenge))

    async def handle_FollowUpQuestionCommand(self, message: FollowUpQuestionCommand):
        follow_up = "Considering your response, how would different climates affect the issue?"
        await self.publish(FollowUpAskedEvent(content=follow_up))

    async def handle_ConcludeDialogueCommand(self, message: ConcludeDialogueCommand):
        conclusion = "Excellent discussion today! You've navigated through the complexities adeptly."
        await self.publish(ConcludeDialogueEvent(content=conclusion))


class StudentActor(BaseActor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.machine = CustomStateMachine(model=self, states=StudentState, initial=StudentState.IDLE)
        self.machine.add_transition(trigger='respond', source=StudentState.IDLE, dest=StudentState.RESPONDING)
        self.machine.add_transition(trigger='consider_follow_up', source=StudentState.RESPONDING, dest=StudentState.CONSIDERING_FOLLOW_UP)
        self.machine.add_transition(trigger='review_feedback', source=[StudentState.CONSIDERING_FOLLOW_UP, StudentState.RESPONDING], dest=StudentState.REVIEWING_FEEDBACK)

    async def handle_AnswerCommand(self, message: AnswerCommand):
        response = "If prices don't change despite the parasites, maybe the impact isn’t as severe."
        await self.publish(AnswerReceivedEvent(content=response))

    async def handle_FollowUpQuestionCommand(self, message: FollowUpQuestionCommand):
        consideration = "Perhaps the environment or climate could be different?"
        await self.publish(FollowUpAskedEvent(content=consideration))


def state_transition_possibilities(actor):
    transition_dict = {}
    for event in actor.machine.events:
        if event.startswith('to_'):
            continue

        # Check if a transition to this state is possible
        transition_possible = getattr(actor.machine.model, f'may_{event}')()
        if transition_possible:
            transition_dict[event] = transition_possible
    return transition_dict


async def main():
    init_dspy()

    actor_system = ActorSystem()
    socrates_actor = await actor_system.actor_of(SocratesActor)
    student_actor = await actor_system.actor_of(StudentActor)

    # Socrates starts the dialogue
    await actor_system.publish(AskQuestionCommand(content="You suggest that the expectation of unchanged prices in Mississippi next year, despite the ongoing issues, weakens the prediction about the parasite's impact outside Mississippi. What led you to this conclusion?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Student responds
    await actor_system.publish(AnswerCommand(content="I thought if prices don't change despite the parasites, maybe the impact isn’t as severe, so the same might happen outside Mississippi."))
    await asyncio.sleep(1)  # Simulate processing time

    # Socrates dives deeper
    await actor_system.publish(FollowUpQuestionCommand(content="Interesting thought. Let's explore further—what factors could potentially influence the severity of the parasite attack outside Mississippi, distinct from what's happening in Mississippi?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Student speculates on environmental factors
    await actor_system.publish(AnswerCommand(content="Maybe the environment or climate could be different?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Socrates pushes for more detailed analysis
    await actor_system.publish(FollowUpQuestionCommand(content="Indeed, environmental differences can greatly affect such situations. Given this, what else might differ in an environment that could influence the impact of parasites?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Student considers biological interactions
    await actor_system.publish(AnswerCommand(content="Could it be something like different plants or maybe animals that interact with the parasites?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Socrates confirms and extends the discussion
    await actor_system.publish(FollowUpQuestionCommand(content="Precisely, interactions with local flora and fauna can indeed play a significant role. How might the presence of certain animals affect the situation with parasites outside Mississippi?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Student acknowledges the role of natural controls
    await actor_system.publish(AnswerCommand(content="If there are animals that eat or control these parasites, they could reduce the damage?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Socrates clarifies and concludes
    await actor_system.publish(FollowUpQuestionCommand(content="Exactly. So, if such a natural control exists outside Mississippi, might that change your view on how the unchanged prices in Mississippi relate to the broader impact of parasites?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Student realizes the implications
    await actor_system.publish(AnswerCommand(content="Yes, I see now. If there are natural controls outside Mississippi, the unchanged prices in Mississippi might not really tell us what could happen elsewhere."))
    await asyncio.sleep(1)  # Simulate processing time

    # Socrates concludes with a reflection
    await actor_system.publish(ConcludeDialogueCommand(content="Very well reasoned! Considering these natural differences, how do you think this knowledge could help us better prepare or adjust agricultural strategies?"))
    await asyncio.sleep(1)  # Simulate processing time

    # Student offers a final insight
    await actor_system.publish(AnswerCommand(content="It seems like understanding and possibly enhancing these natural controls could be important for managing the parasites better."))
    await asyncio.sleep(1)  # Simulate processing time

    # Socrates praises the discussion
    await actor_system.publish(ConcludeDialogueCommand(content="Excellent discussion today! You've adeptly navigated through the complexities of agricultural impacts and the role of local ecological conditions. Your ability to reconsider your initial assumptions when presented with new information shows a strong critical thinking capability. Very well done."))

    print("Dialogue completed.")


if __name__ == '__main__':
    asyncio.run(main())



from dspygen.agents.business_logic import *
from dspygen.agents.order_payload import OrderPayload
from dspygen.mixin.fsm.fsm_mixin import FSMMixin, trigger
from enum import Enum, auto
import logging

# Setup basic configuration for logging
logging.basicConfig(level=logging.INFO, format='== APP == %(asctime)s %(levelname)s: %(message)s',
                    datefmt='%Y-%m-%d %H:%M:%S')


# FSM State Enum
class WorkflowAgentState(Enum):
    INITIALIZING = auto()
    REGISTERING = auto()
    PROCESSING = auto()
    MONITORING = auto()
    COMPLETING = auto()


# FSM Agent Class
class WorkflowFSMAgent(FSMMixin):
    def __init__(self, dapr_adapter, baseInventory):
        super().__init__()
        super().setup_fsm(state_enum=WorkflowAgentState, initial=WorkflowAgentState.INITIALIZING)
        self.dapr_adapter = dapr_adapter
        self.baseInventory = baseInventory
        self.workflow_instance_id = None
        self.order_payload = None
        self.workflow_component = 'workflow_component'
        self.workflow_name = 'order_processing_workflow'

    @trigger(source=WorkflowAgentState.INITIALIZING, dest=WorkflowAgentState.REGISTERING)
    def register_workflow_and_activities(self):
        # self.dapr_adapter.register_workflow(order_processing_workflow)
        self.dapr_adapter.register_activity(notify_activity)
        self.dapr_adapter.register_activity(request_approval_activity)
        self.dapr_adapter.register_activity(verify_inventory_activity)
        self.dapr_adapter.register_activity(process_payment_activity)
        self.dapr_adapter.register_activity(update_inventory_activity)
        # self.dapr_adapter.start()

    @trigger(source=WorkflowAgentState.REGISTERING, dest=WorkflowAgentState.PROCESSING)
    def start_workflow(self, item_name, order_quantity):
        total_cost = int(order_quantity) * self.baseInventory[item_name].per_item_cost
        self.order_payload = OrderPayload(item_name=item_name, quantity=int(order_quantity), total_cost=total_cost)
        start_resp = self.dapr_adapter.start_workflow(self.workflow_component, self.workflow_name,
                                                      input=self.order_payload)
        self.workflow_instance_id = start_resp.instance_id

    @trigger(source=WorkflowAgentState.PROCESSING, dest=WorkflowAgentState.MONITORING)
    def monitor_workflow(self):
        pass

    @trigger(source=WorkflowAgentState.MONITORING, dest=WorkflowAgentState.COMPLETING)
    def complete_workflow(self, result):
        print(f"Workflow completed! Result: {result}")

    # After completing go back to the initial state
    @trigger(source=WorkflowAgentState.COMPLETING, dest=WorkflowAgentState.INITIALIZING)
    def reset_workflow(self):
        print("Resetting workflow agent state")

    def prompt_for_approval(self):
        self.dapr_adapter.raise_workflow_event(self.workflow_instance_id, self.workflow_component, "manager_approval",
                                               {'approval': True})

    def check_workflow_state(self):
        state = self.dapr_adapter.get_workflow(self.workflow_instance_id, self.workflow_component)
        if state.runtime_status in ["Completed", "Failed", "Terminated"]:
            self.trigger('complete_workflow', result=state.runtime_status)


def main():
    from dspygen.utils.dspy_tools import init_ol
    from unittest.mock import MagicMock

    init_ol(max_tokens=3000)
    from dspygen.agents.mock_dapr_adapter import MockDaprClientAdapter
    adapter = MockDaprClientAdapter()
    inventory = {"item1": MagicMock(per_item_cost=100)}  # Setting up a mocked inventory

    agent = WorkflowFSMAgent(adapter, inventory)
    print("Initial state:", agent.state)

    # Simulate the registration and setup of workflows and activities
    agent.prompt("Initialize workflow and activities")
    print("State after initialization:", agent.state)

    # Start the workflow with specific item and quantity
    agent.prompt("Start workflow for item1 with quantity 10", item_name="item1", order_quantity=10)
    print("State after starting the workflow:", agent.state)

    # Mocking a scenario where monitoring might need to reprocess
    # Assuming we have a way to check and decide if reprocessing is necessary
    needs_reprocessing = True  # This should ideally be determined by some condition checks
    if needs_reprocessing:
        agent.prompt("Reprocess workflow")
        print("State after reprocessing:", agent.state)

    # Completing the workflow
    agent.prompt("Complete workflow", result="Success")
    print("State after completion:", agent.state)

    # Resetting the workflow for another cycle or end
    agent.prompt("Reset the workflow")
    print("Final state:", agent.state)


if __name__ == '__main__':
    main()



"""

"""
import dspy
from typer import Typer
from dspygen.utils.dspy_tools import init_dspy


app = Typer()


import dspy

class SelectSingleElement(dspy.Signature):
    """
    Selects a single element from a list of elements that best matches the given prompt.
    """
    elements = dspy.InputField(desc="A list of elements.")
    prompt = dspy.InputField(desc="A prompt used to match the most relevant element.")

    selected_element = dspy.OutputField(desc="The css selector to use with Selenium.")




class GetSelectorModule(dspy.Module):
    """GetSelectorModule"""

    def forward(self, elements, prompt):
        pred = dspy.ChainOfThought(SelectSingleElement)
        result = pred(elements=elements, prompt=prompt).selected_element
        return result


def get_selector_call(elements, prompt):
    get_selector = GetSelectorModule()
    return get_selector.forward(elements=elements, prompt=prompt)


@app.command()
def call(elements, prompt):
    """GetSelectorModule"""
    init_dspy()
    
    print(get_selector_call(elements=elements, prompt=prompt))


from fastapi import APIRouter
router = APIRouter()

@router.post("/get_selector/")
async def get_selector_route(data: dict):
    # Your code generation logic here
    init_dspy()
    
    print(data)
    return get_selector_call(**data)


def main():
    init_dspy()
    element_dicts = """{'type': 'checkbox', 'id': 'vector-main-menu-dropdown-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-vector-main-menu-dropdown', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Main menu'}
{'type': 'checkbox', 'id': 'vector-main-menu-dropdown-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-vector-main-menu-dropdown', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Main menu'}
{'class': 'cdx-text-input__input', 'type': 'search', 'name': 'search', 'placeholder': 'Search Wikipedia', 'aria-label': 'Search Wikipedia', 'autocapitalize': 'sentences', 'title': 'Search Wikipedia [ctrl-option-f]', 'accesskey': 'f', 'id': 'searchInput', 'autocomplete': 'off'}
{'class': 'cdx-text-input__input', 'type': 'search', 'name': 'search', 'placeholder': 'Search Wikipedia', 'aria-label': 'Search Wikipedia', 'autocapitalize': 'sentences', 'title': 'Search Wikipedia [ctrl-option-f]', 'accesskey': 'f', 'id': 'searchInput', 'autocomplete': 'off'}
{'type': 'hidden', 'name': 'title', 'value': 'Special:Search'}
{'type': 'hidden', 'name': 'title', 'value': 'Special:Search'}
{'type': 'checkbox', 'id': 'vector-user-links-dropdown-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-vector-user-links-dropdown', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Personal tools'}
{'type': 'checkbox', 'id': 'vector-user-links-dropdown-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-vector-user-links-dropdown', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Personal tools'}
{'type': 'checkbox', 'id': 'p-variants-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-p-variants', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Change language variant'}
{'type': 'checkbox', 'id': 'p-variants-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-p-variants', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Change language variant'}
{'type': 'checkbox', 'id': 'vector-page-tools-dropdown-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-vector-page-tools-dropdown', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Tools'}
{'type': 'checkbox', 'id': 'vector-page-tools-dropdown-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-vector-page-tools-dropdown', 'class': 'vector-dropdown-checkbox ', 'aria-label': 'Tools'}
{'type': 'checkbox', 'id': 'p-lang-btn-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-p-lang-btn', 'class': 'vector-dropdown-checkbox mw-interlanguage-selector', 'aria-label': 'Go to an article in another language. Available in 48 languages'}
{'type': 'checkbox', 'id': 'p-lang-btn-checkbox', 'role': 'button', 'aria-haspopup': 'true', 'data-event-name': 'ui.dropdown-p-lang-btn', 'class': 'vector-dropdown-checkbox mw-interlanguage-selector', 'aria-label': 'Go to an article in another language. Available in 48 languages'}
"""
    prompt = "search box"
    print(get_selector_call(elements=element_dicts, prompt=prompt))
    

if __name__ == "__main__":
    main()



import dspy

from dspygen.dsl.dsl_pydantic_models import PipelineDSLModel


def _get_language_model_instance(pipeline: PipelineDSLModel, step):
    """
    Get the language model instance for a given step from the top level definition.
    """
    lm_label = step.lm_model
    # Find the lm class within the dspy module. Need to import the class dynamically from the dspy module
    lm_config = next((m for m in pipeline.lm_models if m.label in lm_label), None)

    if lm_config is None:
        raise ValueError(f"Language model with label {lm_label} not found in the pipeline configuration.")

    lm_class = getattr(dspy, lm_config.name)
    lm_inst = lm_class(**lm_config.args)
    return lm_inst


from _ast import stmt
from ast import parse
from dataclasses import dataclass
from typing import Optional

from dspygen.typetemp.environment.typed_environment import TypedEnvironment
from dspygen.typetemp.environment.typed_native_environment import TypedNativeEnvironment
from dspygen.typetemp.template.render_mixin import RenderMixin

_env = TypedEnvironment()
_native_env = TypedNativeEnvironment()


@dataclass
class TypedPythonSource(RenderMixin):
    """Base class for creating templated classes. Uses the jinja2 templating engine
    to render templates. Allows for usage of macros and filters.
    """

    source: str = None  # The string template to be rendered
    use_native: bool = False  # Whether to use NativeEnvironment for rendering
    to: str = None  # The "to" property for rendering destination
    output: str = None  # The rendered output

    def __post_init__(self):
        """After the instance is initialized, set the environment"""
        # Use NativeEnvironment when use_native is True, else use default Environment
        self.env = _native_env if self.use_native else _env

    def __call__(self, **kwargs) -> str:
        return self._render(**kwargs)

    def render_function(self, **kwargs) -> stmt:
        """Renders the function template and returns its AST object.

        :param kwargs: The keyword arguments to be used in rendering
        :return: The AST object of the rendered function
        """
        rendered_func = self._render(**kwargs)
        return parse(rendered_func).body[0]

    def render_class(self, func_tmpls: Optional[list[str]] = None, **kwargs):
        """Renders the class template and returns the compiled class, optionally including methods.

        :param func_tmpls: A list of template strings for functions to be included in the class
        :param kwargs: The keyword arguments to be used in rendering
        :return: The compiled class
        """
        # Render the class
        rendered_cls = self._render(**kwargs)
        class_ast = parse(rendered_cls)

        # If function templates are provided, render and add them to the class
        if func_tmpls:
            for func_tmpl in func_tmpls:
                function_ast = None  #  render_function(func_tmpl, **kwargs)
                class_ast.body[-1].body.append(function_ast)

        # Compile the class AST
        compiled_class_def = compile(class_ast, filename="<ast>", mode="exec")
        class_dict = {}
        exec(compiled_class_def, class_dict)

        # Return the compiled class
        return class_dict[kwargs["class_name"]]



from dspygen.dsl.dsl_pipeline_executor import execute_pipeline
from dspygen.dsl.dsl_pydantic_models import *


def main():
    from dspygen.utils.dspy_tools import init_dspy
    init_dspy()

    module, signature, step = make_step("LinkedIn Newsletter Article Generator", is_first_step=True)

    sig_list = [signature]
    mod_list = [module]
    step_list = [step]

    module, signature, step = make_step(f"Create synthetic critic feedback generator for improvement for {signature.docstring} include critic in the signature name. ", last_sig=signature)

    sig_list.append(signature)
    mod_list.append(module)
    step_list.append(step)

    pipeline = PipelineDSLModel(signatures=sig_list, modules=mod_list, steps=step_list)

    pipeline.to_yaml(f"{signature.name}_pipeline.yaml")

    context = execute_pipeline(f"{signature.name}_pipeline.yaml",
                               {"user_input": "A full stack nextjs DSL pipeline nocode generator"})

    print(context)


def make_step(sig_prompt, last_sig=None, is_first_step=False):
    if last_sig:
        signature = GenSignatureModel.to_inst(f"{sig_prompt}. The inputs should be the outputs {last_sig.outputs}")
    else:
        signature = GenSignatureModel.to_inst(sig_prompt)

    if is_first_step:
        signature.inputs = [signature.inputs[0]]

    module = GenLMModuleModel.to_inst(f"{signature.docstring} convert into a Module. It's name should be similar to {signature.name}")
    module.signature = signature.name

    if is_first_step:
        arg_dict = {signature.inputs[0].name: "{{ user_input }}"}
    else:
        value = "{{" + f" {last_sig.outputs[0].name} " + "}}"
        arg_dict = {signature.inputs[0].name: value}

    step = StepDSLModel(module=module.name, signature=signature.name, args=arg_dict)

    return module, signature, step


if __name__ == '__main__':
    main()



import copy
from typing import Optional, Dict, Any

from dspygen.typetemp.functional import render, render_native
from dspygen.workflow.workflow_models import Workflow, Action, Job
from loguru import logger


def initialize_context(init_ctx: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    """Initializes the workflow context."""
    return copy.deepcopy(init_ctx) if init_ctx else {}


def update_context(context: Dict[str, Any], updates: Dict[str, Any]) -> Dict[str, Any]:
    """Updates the workflow context with new values."""
    # Create a copy of context with only python primitives
    new_context = {k: v for k, v in context.items() if isinstance(v, (int, float, str, bool, list, dict))}

    new_context = copy.deepcopy(new_context)

    new_context.update(updates)

    if '__builtins__' in new_context:
        del new_context['__builtins__']  # Remove builtins from context

    rendered_context = {}
    for arg, value in new_context.items():
        if "{{" in str(value):
            # Render the string value with Jinja2
            rendered_context[arg] = render(value, **new_context)

            # Convert the rendered string to a native Python type
            try:
                rendered_context[arg] = eval(rendered_context[arg])
            except Exception as e:
                logger.error(f"Error converting rendered value to native Python type: {e}")
        else:
            # Non-string values are added to the context unchanged
            rendered_context[arg] = value
    return rendered_context


def evaluate_condition(condition: str, context: Dict[str, Any]) -> bool:
    """Evaluates a condition within the current context."""
    try:
        safe_context = copy.deepcopy(context)
        return eval(condition, {}, safe_context)
    except Exception as e:
        logger.error(f"Error evaluating condition '{condition}': {e}")
        return False


def execute_job(job: Job, context: Dict[str, Any]) -> Dict[str, Any]:
    """Executes all actions within a job."""
    logger.info(f"Executing job: {job.name}")
    job_context = update_context(context, {})  # Isolate context for the job

    for action in job.steps:
        job_context = execute_action(action, job_context)  # Execute each action

    return job_context


def execute_action(action: Action, context: Dict[str, Any]) -> Dict[str, Any]:
    """Executes a single action, updating the context accordingly."""
    logger.info(f"Executing action: {action.name}")

    # Check for conditional execution
    if action.cond and not evaluate_condition(action.cond.expr, context):
        logger.info(f"Condition for action '{action.name}' not met, skipping.")
        return context  # Skip the action if condition not met

    action_context = update_context(context, {})# Isolate context for the action

    if action.code:
        rendered_code = render(action.code, **action_context)
        # Execute action's code, allowing it to modify the action-specific context
        exec(rendered_code, action_context, action_context)
        context = update_context(context, action_context)  # Update global context with changes

    return context


def execute_workflow(workflow: Workflow, init_ctx: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    """Executes all jobs defined in a workflow."""
    logger.info(f"Executing workflow: {workflow.name}")
    global_context = initialize_context(init_ctx)  # Initialize global context

    workflow.process_imports()
    workflow.topological_sort()

    for job in workflow.jobs:
        global_context = execute_job(job, global_context)  # Execute each job

    if '__builtins__' in global_context:
        del global_context['__builtins__']  # Remove builtins from context

    return global_context



import os

from dspygen.utils.pydantic_tools import InstanceMixin


class DataWriter:
    def __init__(self, data, file_path="", write_options=None):
        if write_options is None:
            write_options = {}
        self.file_path = file_path
        self.df = pd.DataFrame(data)
        self.write_options = write_options

    def get_file_path(self):
        context_generator = DataFrameContextGenerator()
        context_string = context_generator.generate_context(self.df)

        inst = FileNameModel.to_inst("Create a filename that fits \n" + context_string)
        return inst.file_name

    def forward(self, **kwargs):
        if not self.file_path:
            self.file_path = self.get_file_path()

        _, file_extension = os.path.splitext(self.file_path)
        file_extension = file_extension.lower()

        write_functions = {
            '.csv': self.df.to_csv,
            # Add more mappings for different file types
        }

        if file_extension in write_functions:
            write_function = write_functions[file_extension]
            try:
                write_function(self.file_path, **self.write_options)
            except Exception as e:
                raise ValueError(f"Failed to write to {self.file_path} due to: {e}")
        else:
            raise ValueError(f"Unsupported file type: {file_extension}")


from pydantic import BaseModel, Field
import pandas as pd
from typing import List, Dict, Any
from io import StringIO


class DataFrameContextGenerator(BaseModel):
    descriptive_stats: bool = True
    dtypes_info: bool = True
    context: str = ""

    class Config:
        arbitrary_types_allowed = True

    def generate_context(self, df) -> str:
        # Convert the input data to a pandas DataFrame

        # Initialize a buffer for DataFrame info
        buffer = StringIO()
        df.info(buf=buffer)
        info_str = buffer.getvalue()

        context_parts = []

        # Optionally include descriptive statistics
        if self.descriptive_stats:
            desc_stats = df.describe().to_string()
            context_parts.append(desc_stats)

        # Optionally include data types information
        if self.dtypes_info:
            dtypes_str = df.dtypes.to_string()
            context_parts.append(dtypes_str)

        # Concatenate all parts to form the complete context
        context = "\n".join(context_parts)
        self.context = context
        return self.context


class FileNameModel(BaseModel, InstanceMixin):
    file_name: str = Field(..., description="Unique CSV filename based on the data provided.")
    extension: str = Field("csv", description="File extension for the output file.")


def main():
    from dspygen.utils.dspy_tools import init_dspy
    init_dspy()
    # Example Usage
    # data = [
    #     {'Date': '2023-01-01', 'Temperature': 22, 'Humidity': 80},
    #     {'Date': '2023-01-02', 'Temperature': 25, 'Humidity': 75},
    #     {'Date': '2023-01-03', 'Temperature': 21, 'Humidity': 85},
    # ]

    data = {
        'Book Title': ['The Great Gatsby', '1984', 'Brave New World', 'The Catcher in the Rye'],
        'Author': ['F. Scott Fitzgerald', 'George Orwell', 'Aldous Huxley', 'J.D. Salinger'],
        'Price': [10.99, 9.99, 8.99, 12.99],
        'Sold Copies': [500, 800, 650, 450]
    }

    # DataWriter(data).forward()
    from dspygen.rm.data_retriever import DataRetriever
    print(DataRetriever("/Users/sac/dev/dspygen/src/dspygen/writer/Book_Title_Author_Price_Sold_Copies.csv").forward())


# Usage example
if __name__ == "__main__":
    main()
#     file_path = 'output_data.csv'
#     data = pd.DataFrame({'id': [1, 2], 'value': ['A', 'B']})
#     writer = DataWriter(file_path, data)
#     writer.write()
#




import yaml
from dspygen.modules.dspygen_module import DGModule

def process_yaml_pipeline(yaml_file):
    with open(yaml_file, 'r') as f:
        config = yaml.safe_load(f)

    result = None

    dg_modules = []

    for module_def in config['modules']:
        module_class = globals()[f"{module_def['module']}DGModule"]  # Get the module class by name
        module_instance = module_class(**module_def.get('args', {}))

        dg_modules.append(module_instance)

    # Pipe the modules together because we need to do __or__ operations
    for i in range(len(dg_modules) - 1):
        dg_modules[i] | dg_modules[i + 1]

    return dg_modules[-1].output


def main():
    from dspygen.utils.dspy_tools import init_dspy
    init_dspy()

    dsl_result = process_yaml_pipeline('pipeline.yaml')
    print(dsl_result)


if __name__ == '__main__':
    main()



from typer.testing import CliRunner

from dspygen.experiments.done.openai_ror_cli import app

runner = CliRunner()

def test_create():
    result = runner.invoke(app, ["create"])
    assert result.exit_code == 0
    assert "This is the create command." in result.output  # Replace with specific expected output

def test_train():
    result = runner.invoke(app, ["train"])
    assert result.exit_code == 0
    assert "This is the train command." in result.output  # Replace with specific expected output

def test_generate():
    result = runner.invoke(app, ["generate"])
    assert result.exit_code == 0
    assert "This is the generate command." in result.output  # Replace with specific expected output

def test_list():
    result = runner.invoke(app, ["list"])
    assert result.exit_code == 0
    assert "This is the list command." in result.output  # Replace with specific expected output

def test_delete():
    result = runner.invoke(app, ["delete"])
    assert result.exit_code == 0
    assert "This is the delete command." in result.output  # Replace with specific expected output

def test_help():
    result = runner.invoke(app, ["help"])
    assert result.exit_code == 0
    assert "This is the help command." in result.output  # Replace with specific expected output




"""
activities.py

This module defines Pydantic models for WS-BPEL 2.0 activities. Activities are the fundamental building blocks of BPEL processes,
defining the business logic and the sequence of operations that are executed within a process.

Included models:
- InvokeActivity: Represents an activity to invoke an operation on a partner Web service.
- ReceiveActivity: Represents an activity that waits for a message from an external partner.
- AssignActivity: Represents an activity for assigning values to variables within the process.
- WaitActivity: Represents an activity that pauses the process execution for a specified duration.
- SequenceActivity: Represents a structured activity that executes a series of activities in a sequential order.
- FlowActivity: Represents a structured activity that allows for concurrent execution of activities.
- PickActivity: Represents a structured activity that waits for one out of several possible events to occur.
- IfActivity: Represents a structured activity that executes one set of activities if a condition is true, and another set if false.
- WhileActivity: Represents a structured activity that repeatedly executes a set of activities while a condition is true.
- RepeatUntilActivity: Represents a structured activity that repeatedly executes a set of activities until a condition becomes true.
- CompensateActivity: Represents an activity that triggers compensation logic for previously executed activities within a scope.
- ThrowActivity: Represents an activity that explicitly throws a fault, causing the process to terminate with a faulted state.
- TerminateActivity: Represents an activity that abruptly terminates the process instance.
- ValidateActivity: Represents an activity that performs validation on a specified variable or data.
- RethrowActivity: Represents an activity that rethrows a caught fault, propagating it to the parent scope for handling.
- CompensateScopeActivity: Represents an activity that triggers compensation logic for the entire scope in which it is defined.
- EmptyActivity: Represents an empty activity, which has no effect on the process execution.
- ScopeActivity: Represents a scoped group of activities with its own local context, including variables, fault handlers, and more.
- ForLoopActivity: Represents a loop activity with a defined start, end, and increment, iterating over a block of activities.
"""

from pydantic import BaseModel, Field
from typing import Optional, List


class Activity(BaseModel):
    """Placeholder for Activities"""


class InvokeActivity(Activity):
    """
    Represents an activity to invoke an operation on a partner Web service.
    Calculus notation: A ::= invoke(L, operation, inputVariable?, outputVariable?) where L is the partnerLink,
    operation is the operation to be invoked, and inputVariable/outputVariable are the optional variables for the invocation.
    """
    id: str = Field(..., description="Unique identifier for the invoke activity.")
    partner_link: str = Field(..., description="Partner link used for the invocation.")
    operation: str = Field(..., description="Operation to invoke on the partner service.")
    input_variable: Optional[str] = Field(None, description="Variable containing data to send.")
    output_variable: Optional[str] = Field(None, description="Variable to store the response from the invocation.")


class ReceiveActivity(Activity):
    """
    Represents an activity that waits for a message from an external partner.
    Calculus notation: A ::= receive(L, operation, variable, createInstance?) where L is the partnerLink,
    operation is the operation associated with the message, variable is where the message should be stored,
    and createInstance is an optional boolean that indicates if this receive activity should create a new process instance.
    """
    id: str = Field(..., description="Unique identifier for the receive activity.")
    partner_link: str = Field(..., description="Partner link expecting a message.")
    operation: str = Field(..., description="Operation associated with the expected message.")
    variable: str = Field(..., description="Variable to store the received message.")
    create_instance: Optional[bool] = Field(None,
                                            description="Whether this activity should initiate a new process instance.")


class AssignActivity(Activity):
    """
    Represents an activity for assigning values to variables within the process.
    Calculus notation: A ::= assign(from, to) where from is an expression or variable value to be assigned,
    and to is the variable or part of a variable that will receive the value.
    """
    id: str = Field(..., description="Unique identifier for the assign activity.")
    from_expression: str = Field(..., description="Expression or variable value to assign from.")
    to_variable: str = Field(..., description="Variable or variable part to assign to.")


class WaitActivity(Activity):
    """
    Represents an activity that pauses the process execution for a specified duration or until a certain condition is met.
    Calculus notation: A ::= wait(for|until) where 'for' is a duration and 'until' is a deadline expression.
    """
    id: str = Field(..., description="Unique identifier for the wait activity.")
    for_: Optional[str] = Field(None, description="Duration to wait for.")
    until: Optional[str] = Field(None, description="Deadline expression to wait until.")


class SequenceActivity(Activity):
    """
    Represents a structured activity that executes a series of activities in a sequential order.
    Calculus notation: A ::= sequence(Activities) where Activities is a list of activities to be executed in order.
    """
    id: str = Field(..., description="Unique identifier for the sequence activity.")
    activities: List[str] = Field(..., description="List of activity identifiers to be executed in sequence.")


class FlowActivity(Activity):
    """
    Represents a structured activity that allows for concurrent execution of activities.
    Calculus notation: A ::= flow(Activities) where Activities is a list of activities that may execute concurrently.
    """
    id: str = Field(..., description="Unique identifier for the flow activity.")
    activities: List[str] = Field(..., description="List of activity identifiers to be executed concurrently.")


class PickActivity(Activity):
    """
    Represents a structured activity that waits for one out of several possible events to occur.
    Calculus notation: A ::= pick(OnMessage | OnAlarm) where OnMessage is a list of message-based branches and
    OnAlarm is a list of time-based branches.
    """
    id: str = Field(..., description="Unique identifier for the pick activity.")
    on_message: Optional[List[str]] = Field(None, description="List of message-based branches.")
    on_alarm: Optional[List[str]] = Field(None, description="List of time-based branches.")


class IfActivity(Activity):
    """
    Represents a structured activity that executes one set of activities if a condition is true, and another set if false.
    Calculus notation: A ::= if(condition) then {Activities} else {Activities}
    """
    id: str = Field(..., description="Unique identifier for the if activity.")
    condition: str = Field(..., description="Condition to evaluate.")
    activities_true: List[str] = Field(..., description="List of activities to execute if the condition is true.")
    activities_false: List[str] = Field(..., description="List of activities to execute if the condition is false.")


class WhileActivity(Activity):
    """
    Represents a structured activity that repeatedly executes a set of activities while a condition is true.
    Calculus notation: A ::= while(condition) {Activities}
    """
    id: str = Field(..., description="Unique identifier for the while activity.")
    condition: str = Field(..., description="Condition to evaluate for each iteration.")
    activities: List[str] = Field(..., description="List of activities to execute in each iteration.")


class RepeatUntilActivity(Activity):
    """
    Represents a structured activity that repeatedly executes a set of activities until a condition becomes true.
    Calculus notation: A ::= repeat {Activities} until (condition)
    """
    id: str = Field(..., description="Unique identifier for the repeat-until activity.")
    activities: List[str] = Field(..., description="List of activities to execute in each iteration.")
    condition: str = Field(..., description="Condition to evaluate after each iteration.")


class CompensateActivity(Activity):
    """
    Represents an activity that triggers compensation logic for previously executed activities within a scope.
    Calculus notation: A ::= compensate
    """
    id: str = Field(..., description="Unique identifier for the compensate activity.")


class ThrowActivity(Activity):
    """
    Represents an activity that explicitly throws a fault, causing the process to terminate with a faulted state.
    Calculus notation: A ::= throw faultName
    """
    id: str = Field(..., description="Unique identifier for the throw activity.")
    fault_name: str = Field(..., description="Name of the fault to throw.")


class TerminateActivity(Activity):
    """
    Represents an activity that abruptly terminates the process instance.
    Calculus notation: A ::= terminate
    """
    id: str = Field(..., description="Unique identifier for the terminate activity.")


class ValidateActivity(Activity):
    """
    Represents an activity that performs validation on a specified variable or data.
    Calculus notation: A ::= validate(variable)
    """
    id: str = Field(..., description="Unique identifier for the validate activity.")
    variable: str = Field(..., description="Variable or data to be validated.")


class RethrowActivity(Activity):
    """
    Represents an activity that rethrows a caught fault, propagating it to the parent scope for handling.
    Calculus notation: A ::= rethrow
    """
    id: str = Field(..., description="Unique identifier for the rethrow activity.")


class CompensateScopeActivity(Activity):
    """
    Represents an activity that triggers compensation logic for the entire scope in which it is defined.
    Calculus notation: A ::= compensateScope
    """
    id: str = Field(..., description="Unique identifier for the compensateScope activity.")


class EmptyActivity(Activity):
    """
    Represents an empty activity, which has no effect on the process execution.
    Calculus notation: A ::= empty
    """
    id: str = Field(..., description="Unique identifier for the empty activity.")


class ScopeActivity(Activity):
    """
    Represents a scoped group of activities with its own local context, including variables, fault handlers, and more.
    Calculus notation: A ::= scope {Activities, Variables, FaultHandlers, CompensationHandlers, EventHandlers}
    """
    id: str = Field(..., description="Unique identifier for the scope activity.")
    activities: List[str] = Field(..., description="List of activity identifiers contained within the scope.")
    variables: List[str] = Field([], description="List of local variable identifiers declared in the scope.")
    fault_handlers: List[str] = Field([], description="List of fault handler identifiers associated with the scope.")
    compensation_handlers: List[str] = Field([], description="List of compensation handler identifiers for the scope.")
    event_handlers: List[str] = Field([], description="List of event handler identifiers for the scope.")


class ForLoopActivity(Activity):
    """
    Represents a loop activity with a defined start, end, and increment, iterating over a block of activities.
    Note: This is an extension for convenience and not part of the standard BPEL model.
    """
    id: str = Field(..., description="Unique identifier for the for loop activity.")
    start_expression: str = Field(..., description="Expression defining the loop's start value.")
    end_expression: str = Field(..., description="Expression defining the loop's end value.")
    increment_expression: str = Field(..., description="Expression defining the loop's increment.")
    activities: List[str] = Field(...,
                                  description="List of activity identifiers to be executed in each loop iteration.")



from loguru import logger

from dspygen.modules.python_source_code_module import python_source_code_call
from dspygen.rm.chatgpt_chromadb_retriever import ChatGPTChromaDBRetriever
from dspygen.utils.dspy_tools import init_dspy
from dspygen.lm.groq_lm import Groq


def main():
    init_dspy(lm_class=Groq, model="mixtral-8x7b-32768")

    retriever = ChatGPTChromaDBRetriever()
    query = "Revenue Operations Automation"
    matched_conversations = retriever.forward(query, k=10)
    # print(count_tokens(str(matched_conversations) + "\nI want a DSPy module that generates Python source code."))
    for conversation in matched_conversations:
        logger.info(conversation)

    logger.info(python_source_code_call(str(matched_conversations)))


if __name__ == '__main__':
    main()



"""Generate DSPy module"""
from importlib import import_module

import os

import inflection
import typer

from dspygen.dsl.dsl_pipeline_executor import execute_pipeline
from dspygen.modules.gen_dspy_module import  DSPyModuleTemplate, SignatureDspyModuleModule
from dspygen.utils.cli_tools import chatbot
from dspygen.utils.dspy_tools import init_dspy
from dspygen.utils.file_tools import dspy_modules_dir, source_dir, get_source

app = typer.Typer(help="Generate DSPy Modules or call exist ones.")


@app.command(name="new")
def new_module(
        class_name: str = typer.Option(..., "--class-name", "-cn", help="The name of the module class"),
        inputs: str = typer.Option(None, "--inputs", "-i", help="A comma-separated list of input names"),
        output: str = typer.Option(None, "--output", "-o", help="Output name for the module"),
):
    """Generate a new dspy.Module. Inputs and output will be static."""
    if len(inputs) == 0 or output is None:
        raise ValueError("Please provide a signature or input and output.")

    mdl = DSPyModuleTemplate(class_name=class_name, inputs=inputs.split(','), output=output)

    source = SignatureDspyModuleModule().forward(mdl)

    file_name = f"{inflection.underscore(class_name)}_module.py"

    with open(dspy_modules_dir() / file_name, "w") as file:
        file.write(source)

    print(source)

    print(f"Module saved to {dspy_modules_dir() / file_name}")


def load_commands(directory: str = "modules"):
    subcommands_dir = source_dir() / directory

    for filename in os.listdir(subcommands_dir):
        if filename.endswith("_module.py"):
            module_name = f'{__name__.split(".")[0]}.{directory}.{filename[:-3]}'
            module = import_module(module_name)
            if hasattr(module, "app"):
                app.add_typer(module.app, name=filename[:-10])


def main():
    inputs = ["input1", "input2", "input3"]
    output = "output"
    print()


if __name__ == "__main__":
    main()
else:
    load_commands()


@app.command("help")
def cli_help(question: str):
    """Answers the user questions with a helpful chatbot."""
    chatbot(question, get_source(__file__))





"""

"""
import dspy
from dspygen.utils.dspy_tools import init_dspy


import dspy

class GenerateStreamlitComponents(dspy.Signature):
    """
    Generate Streamlit components source code based on project and page information.
    Do not return ```
    """
    project = dspy.InputField(desc="Information or context about the project.")
    page = dspy.InputField(desc="Specific details or requirements for the Streamlit page.")

    streamlit_components_source_code = dspy.OutputField(desc="Generated source code for Streamlit components to make the page functional.",
                                                        prefix="```python\n")



class StreamlitBotModule(dspy.Module):
    """StreamlitBotModule"""

    def forward(self, project, page):
        page += "\nProvide detailed streamlit components that would make the page functional"
        pred = dspy.ChainOfThought(GenerateStreamlitComponents)
        result = pred(project=project, page=page).streamlit_components_source_code
        return result


from typer import Typer
app = Typer()


@app.command()
def call(project, page):
    """StreamlitBotModule"""
    init_dspy()

    print(streamlit_bot_call(project=project, page=page))



def streamlit_bot_call(project, page):
    streamlit_bot = StreamlitBotModule()
    return streamlit_bot.forward(project=project, page=page)



def main():
    init_dspy()
    project = ""
    page = ""
    print(streamlit_bot_call(project=project, page=page))



from fastapi import APIRouter
router = APIRouter()

@router.post("/streamlit_bot/")
async def streamlit_bot_route(data: dict):
    # Your code generation logic here
    init_dspy()

    print(data)
    return streamlit_bot_call(**data)



"""
import streamlit as st


# Streamlit form and display
st.title("StreamlitBotModule Generator")
project = st.text_input("Enter project")
page = st.text_input("Enter page")

if st.button("Submit StreamlitBotModule"):
    init_dspy()

    result = streamlit_bot_call(project=project, page=page)
    st.write(result)
"""

if __name__ == "__main__":
    main()



from pydantic import BaseModel, ValidationError
import json

class DecisionInput(BaseModel):
    income: float
    credit_score: int
    loan_amount: float

class DecisionOutput(BaseModel):
    loan_approval: str
    maximum_loan_amount: float

class Decision(BaseModel):
    inputs: DecisionInput
    outputs: DecisionOutput

def populate_model(model: BaseModel, data: dict, depth=0):
    field_values = {}
    for field_name, field_type in model.__annotations__.items():
        if issubclass(field_type, BaseModel):
            # Recurse into submodels
            field_values[field_name] = populate_model(field_type, data.get(field_name, {}), depth+1)
        else:
            # Directly assign data if available
            field_values[field_name] = data.get(field_name)

    try:
        # Create model instance with populated data
        return model(**field_values)
    except ValidationError as ve:
        print(f"Validation error at depth {depth}: {ve}")
        return None

# Example usage
data = {
    "inputs": {"income": 5000.00, "credit_score": 720, "loan_amount": 15000},
    "outputs": {"loan_approval": "Approved", "maximum_loan_amount": 15000}
}

decision_model = populate_model(Decision, data)
print(decision_model)



from enum import Enum, auto
from dspygen.mixin.fsm.fsm_mixin import FSMMixin, trigger
from dspygen.utils.dspy_tools import init_ol
from dspygen.modules.fsm_trigger_module import fsm_trigger_call


class AnalyticsAgentState(Enum):
    IDLE = auto()
    ANALYZING = auto()
    REPORTING = auto()
    HANDLING_ERRORS = auto()
    COMPLETING_TASK = auto()


class AnalyticsAgent(FSMMixin):
    def __init__(self):
        super().setup_fsm(AnalyticsAgentState, initial=AnalyticsAgentState.IDLE)

    @trigger(source=AnalyticsAgentState.IDLE, dest=AnalyticsAgentState.ANALYZING)
    def start_analysis(self):
        print("Starting data analysis.")

    @trigger(source=AnalyticsAgentState.ANALYZING, dest=AnalyticsAgentState.REPORTING)
    def generate_report(self):
        print("Generating analysis report.")

    @trigger(source=[AnalyticsAgentState.ANALYZING, AnalyticsAgentState.REPORTING],
             dest=AnalyticsAgentState.HANDLING_ERRORS)
    def handle_error(self):
        print("Handling error in data analysis.")

    @trigger(source=AnalyticsAgentState.HANDLING_ERRORS, dest=AnalyticsAgentState.IDLE)
    def resolve_error(self):
        print("Error resolved, returning to idle.")

    @trigger(source=[AnalyticsAgentState.ANALYZING, AnalyticsAgentState.REPORTING],
             dest=AnalyticsAgentState.COMPLETING_TASK)
    def complete_task(self):
        print("Data analysis task completed.")


def main():
    init_ol(max_tokens=3000)  # Initialize any required DSPy settings
    agent = AnalyticsAgent()
    print("Initial state:", agent.state)

    # Simulate voice commands
    fsm_trigger_call("I want you to start analyzing the data", agent)
    fsm_trigger_call("Please generate the report now", agent)
    fsm_trigger_call("There is an issue, handle the error", agent)
    fsm_trigger_call("Now resolve the error", agent)
    fsm_trigger_call("Complete the data analysis task", agent)

    print("Final state:", agent.state)


if __name__ == "__main__":
    main()




from enum import Enum, auto
from dspygen.mixin.fsm.fsm_mixin import FSMMixin, trigger


class TrafficLightState(Enum):
    GREEN = auto()
    YELLOW = auto()
    RED = auto()


class TrafficLight(FSMMixin):
    def __init__(self):
        super().setup_fsm(TrafficLightState)
        # Initialize state-specific attributes if any

    @trigger(
        source=TrafficLightState.GREEN,
        dest=TrafficLightState.YELLOW,
        conditions=None,
        unless=None,
        before=None,
        after=None,
        prepare=None
    )
    def slow_down(self):
        print("Transitioning from green to yellow")
        
    @trigger(
        source=TrafficLightState.YELLOW,
        dest=TrafficLightState.RED,
        conditions=None,
        unless=None,
        before=None,
        after=None,
        prepare=None
    )
    def stop(self):
        print("Transitioning from yellow to red")
        
    @trigger(
        source=TrafficLightState.RED,
        dest=TrafficLightState.GREEN,
        conditions=None,
        unless=None,
        before=None,
        after=None,
        prepare=None
    )
    def go(self):
        print("Transitioning from red to green")
        
    def log_transition(self):
        print("Logging transition.")

    def celebrate_red(self):
        print("Celebrating red light!")


def main():
    """Main function"""
    tl = TrafficLight()
    tl.slow_down()  # Transition from GREEN to YELLOW
    tl.stop()  # Transition from YELLOW to RED

    assert tl.state == TrafficLightState.RED.name




if __name__ == '__main__':
    main()    


from pydantic import BaseModel, Field
from typing import List, Optional

from dspygen.utils.yaml_tools import YAMLMixin, uuid_factory


class Flow(BaseModel):
    """
    Represents a flow within a BPMN process, defining the sequence or message flow between elements.
    """
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the flow.")
    sourceRef: str = Field(..., description="The source element of the flow.")
    targetRef: str = Field(..., description="The target element the flow points to.")
    condition: Optional[str] = Field(None, description="The condition that determines whether the flow is taken.")


class Task(BaseModel):
    """
    Represents a task within a BPMN process. A task is a unit of work within a process.
    """
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the task.")
    name: str = Field(..., description="The name of the task.")
    type: str = Field(..., description="The type of the task, e.g., 'serviceTask', 'userTask'.")
    properties: Optional[dict] = Field(default=None,
                                       description="Additional properties of the task, defined as a dictionary.")


class Process(BaseModel):
    """
    Represents a BPMN process, which is a collection of tasks and flows that define the workflow.
    """
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the process.")
    name: str = Field(..., description="The name of the process.")
    tasks: List[Task] = Field(..., description="A list of tasks within the process.")
    flows: List[Flow] = Field(...,
                              description="A list of flows that define the sequence and message flows within the process.")


class BPMN(BaseModel, YAMLMixin):
    """
    Defines a BPMN model, which may contain one or more processes. This model includes methods for parsing from and dumping to YAML.
    """
    processes: List[Process] = Field(..., description="A list of processes defined in the BPMN model.")


class Sentry(BaseModel):
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the sentry.")
    name: str
    condition: str


class PlanItem(BaseModel):
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the plan item.")
    name: str
    type: str
    role: str
    description: str
    entryCriterion: Optional[str] = None


class Stage(BaseModel):
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the stage.")
    name: str
    tasks: List[PlanItem]


class Case(BaseModel):
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the case.")
    name: str
    roles: List[str]
    stages: List[Stage]
    planItems: List[PlanItem]
    sentries: List[Sentry]


class CMMN(BaseModel, YAMLMixin):
    cases: List[Case]


class Input(BaseModel):
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the input.")
    label: str
    inputExpression: str
    inputValues: List[str]


class Output(BaseModel):
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the output.")
    label: str
    outputValues: List[str]


class Rule(BaseModel):
    inputEntries: List[str] = Field(
        ...,
        description="The conditions under which the rule applies.",
        example=["Credit Score >= 700", "Income >= 3000", "Loan Amount <= 50000"]  # Adding example directly to the field
    )
    outputEntries: List[str] = Field(
        ...,
        description="The outcomes or actions taken when conditions are met.",
        example=["Loan Approved", "Interest Rate: 3.5%", "Maximum Loan Amount: 50000"]  # Adding example directly to the field
    )


class DecisionTable(BaseModel):
    inputs: List[Input]
    outputs: List[Output]
    rules: List[Rule]


class Decision(BaseModel):
    id: str = Field(default_factory=uuid_factory, description="The unique identifier for the decision.")
    name: str
    decisionTable: DecisionTable


class DMN(BaseModel, YAMLMixin):
    definitions: dict[str, Decision]


def main():
    """Main function"""
    from dspygen.utils.dspy_tools import init_ol
    # init_ol()
    from dspygen.utils.dspy_tools import init_dspy
    init_dspy()
    print(DMN.model_json_schema())
    from dspygen.modules.prompt_to_json_module import instance
    DMN.model_validate_json()
    print(instance(DMN.model_json_schema(), ))


if __name__ == '__main__':
    main()



import dspy


class WebRetriever(dspy.Retrieve):
    def __init__(self, source, **kwargs):
        super().__init__()

        self.source = source

    def forward(self, query, **kwargs):
        return None


def main():
    rm = WebRetriever(source="<html><form><button type='submit'/></form></html>")
    print(rm.forward(query="What is the css selector for the submit button?"))


if __name__ == '__main__':
    main()



import dspy
from pydantic import BaseModel

from dspygen.typetemp.functional import render


class AssertionModel(BaseModel):
    label: str
    constraint: str
    message: str
    args: dict


class AssertionExecutor:
    def __init__(self, assertion_model: AssertionModel):
        self.assertion_model = assertion_model

    def execute_assertion(self, result):
        self.assertion_model.args["result"] = result
        constraint = render(self.assertion_model.constraint, **self.assertion_model.args)
        message = render(self.assertion_model.message, **self.assertion_model.args)

        dspy.Assert(eval(constraint), message)


# Example usage
assertion_data = {
    "label": "AssertLen",
    "constraint": "{{ result | length }} >= {{ min_length }}",
    "message": "The length of the {{ result }} should be {{ min_length }} minimum but was {{ result | length }}",
    "args": {"min_length": 10}
}


def main():
    assertion_model = AssertionModel(**assertion_data)
    executor = AssertionExecutor(assertion_model)

    # Assuming result is some data you want to test
    result = "Hello, world!"

    try:
        executor.execute_assertion(result)
    except AssertionError as e:
        print(e)


if __name__ == '__main__':
    main()



import json
import tiktoken
import anyio
import yaml

from pathlib import Path
from fnmatch import fnmatch

import tempfile
from contextlib import contextmanager


def extract_code(text: str) -> str:
    # Use a regular expression to find code blocks enclosed in triple backticks.
    text_code = re.findall(r"```([\s\S]+?)```", text)

    if not text_code:
        return text

    # Concatenate all the code blocks together with double newline separators.
    concatenated_code = "\n\n".join(
        [code[code.index("\n") + 1:] for code in text_code]
    )

    return concatenated_code


import re


def slugify(text):
    # Convert the text to lowercase
    text = text.lower()

    text = text.replace(" ", "-")

    # Replace spaces with hyphens and remove other non-alphanumeric characters
    text = re.sub(r"[^a-z0-9-]", "", text)

    # Replace multiple consecutive hyphens with a single hyphen
    text = re.sub(r"[-]+", "-", text)

    # Remove leading and trailing hyphens
    text = text.strip("-")

    return text


def find_project_root(current_path: Path | str = Path(__file__)) -> Path:
    """
    Traverse up from the current path to find the project root marker.
    """
    if isinstance(current_path, str):
        current_path = Path(current_path)

    # Define the name of your project root marker
    marker = "pyproject.toml"
    # Start from the current file's directory
    current_dir = current_path.parent
    while current_dir != current_dir.parent:  # stop at the root of the filesystem
        if (current_dir / marker).exists():
            return current_dir
        current_dir = current_dir.parent
    raise FileNotFoundError(f"Project root marker '{marker}' not found.")


def project_dir() -> Path:
    return Path(__file__).parent.parent.parent.parent


def data_dir(filename="") -> Path:
    return project_dir() / "data" / filename


def source_dir(file_name="") -> Path:
    return Path(__file__).parent.parent / file_name


def subcommands_dir(file_name="") -> Path:
    return source_dir() / "subcommands" / file_name


def dspy_modules_dir(file_name="") -> Path:
    return source_dir() / "modules" / file_name


def signatures_dir(file_name="") -> Path:
    return source_dir() / "signatures" / file_name


def lm_dir(file_name="") -> Path:
    return source_dir() / "lm" / file_name


def rm_dir(file_name="") -> Path:
    return source_dir() / "rm" / file_name


def templates_dir(file_name="") -> Path:
    return source_dir() / "templates" / file_name


def pages_dir(file_name="") -> Path:
    return source_dir() / "pages" / file_name


def dsl_dir(file_name="") -> Path:
    return source_dir() / "dsl" / file_name


def get_source(filename):
    # Read the source code from the file
    with open(filename, 'r') as file:
        source_code = file.read().replace(" ", "")

    return source_code


def is_dspygen():
    return "dspygen" in get_source(__file__)


def count_tokens(text: str, model: str = "gpt-4") -> int:
    enc = tiktoken.encoding_for_model("gpt-4")
    return len(enc.encode(text))


async def read(filename, to_type=None):
    async with await anyio.open_file(filename, mode="r") as f:
        contents = await f.read()
    if to_type == "dict":
        if filename.endswith(".yaml") or filename.endswith(".yml"):
            contents = yaml.safe_load(contents)
        elif filename.endswith(".json"):
            contents = json.loads(contents)
    return contents


async def write(
        contents=None,
        *,
        filename=None,
        mode="w+",
        extension="txt",
        time_stamp=False,
        path="",
):
    # if extension == "yaml" or extension == "yml":
    #     contents = yaml.dump(
    #         contents, default_style="", default_flow_style=False, width=1000
    #     )
    # elif extension == "json":
    #     contents = json.dumps(contents)

    async with await anyio.open_file(path + filename, mode=mode) as f:
        await f.write(contents)
    return filename


@contextmanager
def tmp_file(content, mode='w+', delete=True):
    """
    Creates a temp file with content, yielding its path.

    Args:
        content (str): Content to write to the file.
        mode (str): File mode ('w+' for text, 'wb+' for binary).
        delete (bool): Delete file on exit (default True).

    Yields:
        str: The path to the temporary file.
    """
    with tempfile.NamedTemporaryFile(mode=mode, delete=delete) as tmp:
        # Write the content based on the mode
        if 'b' in mode:
            # Ensure content is bytes if in binary mode
            if isinstance(content, str):
                content = content.encode()  # Convert to bytes
            tmp.write(content)
        else:
            # Ensure content is str if in text mode
            if isinstance(content, bytes):
                content = content.decode()  # Convert to str
            tmp.write(content)

        tmp.flush()  # Ensure content is written
        yield tmp.name  # Provide file path


# Test the function
def main():
    print(count_tokens(get_source(__file__)))





def parse_gitignore(gitignore_path):
    if not gitignore_path.exists():
        return set()

    with gitignore_path.open("r", encoding="utf-8") as file:
        patterns = set(line.strip() for line in file if line.strip() and not line.startswith("#"))
    patterns.add(".git")  # Always ignore .git directory
    return patterns


def is_ignored(file_path, ignore_patterns):
    relative_path = file_path.relative_to(file_path.parents[len(ignore_patterns) - 1])
    return any(match_gitignore_pattern(relative_path, pattern) for pattern in ignore_patterns)


def match_gitignore_pattern(relative_path, pattern):
    if pattern.startswith("/"):
        if fnmatch(str(relative_path), pattern[1:]) or fnmatch(str(relative_path.parent), pattern[1:]):
            return True
    else:
        if any(fnmatch(str(path), pattern) for path in [relative_path, *relative_path.parents]):
            return True
    return False


def is_binary(file_path):
    try:
        with open(file_path, "rb") as file:
            return b"\x00" in file.read(1024)
    except IOError:
        return False


def find_gitignore(start_path):
    """
    Walks up the directory tree from the start_path to find a .gitignore file.

    :param start_path: The starting directory path as a Path object or string.
    :return: The path to the first .gitignore file found, or None if not found.
    """
    current_path = Path(start_path).resolve()  # Ensure we have an absolute path
    for parent in [current_path, *current_path.parents]:
        gitignore_path = parent / ".gitignore"
        if gitignore_path.exists():
            return gitignore_path
    return None


if __name__ == "__main__":
    main()



import json
import uuid
from dataclasses import dataclass
from pathlib import Path

import openai
from dateutil import parser
from icontract import require, ensure
from pydantic import BaseModel
from slugify import slugify
from sqlmodel import Field, SQLModel, Relationship, select
from datetime import datetime
from typing import Optional, List, NamedTuple

from tzlocal import get_localzone

from dspygen.experiments.rfc5545.ical_db_session import get_session
from dspygen.utils.crud_tools import add_model, update_model, delete_model, get_model
from dspygen.utils.date_tools import parse_datetime


class OptionConfig(BaseModel):
    model: Optional[str]
    input_file: Optional[str]
    output_file: Optional[str]
    prompt: Optional[str]
    verbose: Optional[bool]
    auto_save: Optional[bool]
    paste_from_clipboard: Optional[bool]
    file_extension: Optional[str]
    append_to_output: Optional[bool]
    response: Optional[str]
    max_tokens: Optional[int]


class Calendar(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    prodid: str = Field(description="Product Identifier")
    version: str = Field(description="Version of the calendar")
    calscale: Optional[str] = Field(description="Calendar scale used")
    method: Optional[str] = Field(description="Method used in the calendar")
    events: List["Event"] = Relationship(back_populates="calendar")
    todos: List["Todo"] = Relationship(back_populates="calendar")
    journals: List["Journal"] = Relationship(back_populates="calendar")


class Event(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    uid: str = Field(description="Globally unique identifier")
    dtstamp: datetime = Field(description="Date/time stamp")
    dtstart: datetime = Field(description="Start date/time of the event")
    dtend: Optional[datetime] = Field(description="End date/time of the event")
    duration: Optional[str] = Field(description="Duration of the event")
    summary: Optional[str] = Field(description="Summary of the event")
    description: Optional[str] = Field(description="Full description of the event")
    location: Optional[str] = Field(description="Location of the event")
    calendar_id: Optional[int] = Field(default=None, foreign_key="calendar.id")
    calendar: Optional[Calendar] = Relationship(back_populates="events")
    alarms: List["Alarm"] = Relationship(back_populates="event")

    def to_ics(self, file_path: Optional[str] = None) -> str:
        """
        Converts the event to iCalendar format with the user's local timezone.
        """
        # Get the user's local timezone
        local_tz = get_localzone()

        # Convert event start and end times to the user's local timezone
        dtstart_local = self.dtstart.astimezone(local_tz)
        dtend_local = self.dtend.astimezone(local_tz) if self.dtend else None

        # Start the calendar
        ics_str = "BEGIN:VCALENDAR\n"
        ics_str += "VERSION:2.0\n"
        ics_str += "PRODID:-//YourCompany//YourProduct//EN\n"  # Example Product ID

        # Event details
        ics_str += "BEGIN:VEVENT\n"
        ics_str += f"UID:{self.uid}\n"
        ics_str += f'DTSTAMP:{self.dtstamp.strftime("%Y%m%dT%H%M%SZ")}\n'
        ics_str += f'DTSTART:{dtstart_local.strftime("%Y%m%dT%H%M%S")}\n'
        if dtend_local:
            ics_str += f'DTEND:{dtend_local.strftime("%Y%m%dT%H%M%S")}\n'
        if self.summary:
            ics_str += f"SUMMARY:{self.summary}\n"
        if self.description:
            ics_str += f"DESCRIPTION:{self.description}\n"
        if self.location:
            ics_str += f"LOCATION:{self.location}\n"
        ics_str += "END:VEVENT\n"

        # End the calendar
        ics_str += "END:VCALENDAR\n"

        # If path is a directory, save the ICS content to a file in that directory
        # the file name will be the event slugged summary and dtstart with a .ics extension
        if file_path and Path(file_path).is_dir():
            file_name = (
                f"{slugify(self.summary)}_{self.dtstart.strftime('%Y%m%dT%H%M%SZ')}.ics"
            )
            file_path = Path(file_path) / file_name
            with open(file_path, "w") as f:
                f.write(ics_str)
        elif file_path:
            with open(file_path, "w") as f:
                f.write(ics_str)

        return ics_str

    def __repr__(self):
        return (
            f"Event(id={self.id}, summary='{self.summary}', dtstart={self.dtstart}, "
            f"dtend={self.dtend}, duration={self.duration}, description='{self.description}', "
            f"location='{self.location}')"
        )

    def __str__(self):
        dtstart_str = self.dtstart.strftime("%I:%M%p on %A, %B %d, %Y")
        dtend_str = (
            self.dtend.strftime("%I:%M%p on %A, %B %d, %Y") if self.dtend else ""
        )
        return (
            f"Summary:\t{self.summary}\n"
            f"Start:\t\t{dtstart_str}\n"
            f"End:\t\t{dtend_str}\n"
            f"Description:\t{self.description}\n"
            f"Location:\t{self.location}"
        )

    @staticmethod
    @require(lambda dtstart: parser.parse(dtstart))
    @require(lambda dtend: dtend is None or parser.parse(dtend))
    @require(lambda duration: duration is None or isinstance(duration, str))
    @require(lambda summary: summary is None or isinstance(summary, str))
    @require(lambda description: description is None or isinstance(description, str))
    @require(lambda location: location is None or isinstance(location, str))
    def create(
        dtstart: str,
        dtend: str = None,
        duration: str = None,
        summary: str = None,
        description: str = None,
        location: str = None,
    ) -> "Event":
        uid = uuid.uuid4()
        uid_str = str(uid)

        dtstamp = datetime.utcnow()

        event = Event(
            uid=uid_str,
            dtstamp=dtstamp,
            dtstart=parser.parse(dtstart),
            dtend=parser.parse(dtend) if dtend else None,
            duration=duration,
            summary=summary,
            description=description,
            location=location,
        )

        add_model(event)
        return event

    @staticmethod
    @require(lambda event_id: isinstance(event_id, int))
    @require(lambda dtstart: dtstart is None or parser.parse(dtstart))
    @require(lambda dtend: dtend is None or parser.parse(dtend))
    @require(lambda duration: duration is None or isinstance(duration, str))
    @require(lambda summary: summary is None or isinstance(summary, str))
    @require(lambda description: description is None or isinstance(description, str))
    @require(lambda location: location is None or isinstance(location, str))
    def update(
        event_id: int,
        dtstart: str = None,
        dtend: str = None,
        duration: str = None,
        summary: str = None,
        description: str = None,
        location: str = None,
    ):
        with update_model(Event, event_id) as event:
            event.dtstamp = datetime.utcnow()
            if dtstart:
                event.dtstart = parser.parse(dtstart)
            if dtend:
                event.dtend = parser.parse(dtend) if dtend else None
            if duration:
                event.duration = duration
            if summary:
                event.summary = summary
            if description:
                event.description = description
            if location:
                event.location = location

    @staticmethod
    @require(lambda event_id: isinstance(event_id, int))
    @ensure(lambda result, event_id: result is None)
    def delete(event_id: int) -> None:
        delete_model(Event, event_id)

    @staticmethod
    @require(lambda event_id: isinstance(event_id, int))
    @ensure(lambda result, event_id: result.id == event_id)
    def read(event_id: int) -> "Event":
        return get_model(Event, event_id)

    @staticmethod
    def most_recent() -> "Event":
        return get_session().exec(select(Event).order_by(Event.dtstamp.asc())).first()

    @staticmethod
    def get_by_page(
        page: int = 0,
        per_page: int = 10,
        sort: str = "dtstart",
        asc: bool = True,
        include_past: bool = False,
    ) -> List["Event"]:
        """Get events by page sorted by"""
        order_by = getattr(Event, sort).asc() if asc else getattr(Event, sort).desc()

        query = select(Event).order_by(order_by)

        if not include_past:
            # Add a filter to exclude past events
            query = query.filter(Event.dtstart >= datetime.now())

        return get_session().exec(query.offset(page * per_page).limit(per_page)).all()

    # @staticmethod
    # def query(text: str, where: dict = None, n_results: int = 10) -> List["Event"]:
    # TODO: Reimplement with
    #     """Query events"""
    #     results = get_mem_store().query(
    #         collection_id="Event_collection",
    #         query=text,
    #         where=where,
    #         n_results=n_results,
    #     )
    #     result_ids = [json.loads(r)["id"] for r in results["documents"][0]]
    #     models = get_session().exec(select(Event).where(Event.id.in_(result_ids))).all()
    #     # sort models by result_ids
    #     return sorted(models, key=lambda m: result_ids.index(m.id))


class Todo(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    uid: str = Field(description="Globally unique identifier")
    dtstamp: datetime = Field(description="Date/time stamp")
    dtstart: Optional[datetime] = Field(description="Start date/time of the todo")
    due: Optional[datetime] = Field(description="Due date/time of the todo")
    summary: Optional[str] = Field(description="Summary of the todo")
    description: Optional[str] = Field(description="Description of the todo")
    completed: Optional[datetime] = Field(description="Completion date/time")
    calendar_id: Optional[int] = Field(default=None, foreign_key="calendar.id")
    calendar: Optional[Calendar] = Relationship(back_populates="todos")

    @staticmethod
    @require(lambda summary: summary is isinstance(summary, str))
    @ensure(lambda result: result.id is not None)
    def create(
            summary: str,
            dtstart: str | datetime = None,
            due: str | datetime = None,
            description: str = None,
            completed: str | datetime = None,
            calendar_id: Optional[int] = None
    ) -> "Todo":
        uid = uuid.uuid4()
        uid_str = str(uid)

        dtstamp = datetime.utcnow()

        new_todo = Todo(
            uid=uid_str,
            dtstamp=dtstamp,
            dtstart=parse_datetime(dtstart),
            due=parse_datetime(due),
            summary=summary,
            description=description,
            completed=parse_datetime(completed),
            calendar_id=calendar_id
        )

        add_model(new_todo)

        return new_todo

    @staticmethod
    @require(lambda todo_id: isinstance(todo_id, int))
    def update(
            todo_id: int,
            summary: str,
            dtstart: str | datetime = None,
            due: str | datetime = None,
            description: str = None,
            completed: str | datetime = None,
            calendar_id: Optional[int] = None
    ):
        with update_model(Todo, todo_id) as todo:
            todo.dtstamp = datetime.utcnow()
            todo.dtstart = parse_datetime(dtstart)
            todo.due = parse_datetime(due)
            todo.summary = summary
            todo.description = description
            todo.completed = parse_datetime(completed)
            todo.calendar_id = calendar_id

    @staticmethod
    @require(lambda todo_id: isinstance(todo_id, int))
    def delete(todo_id: int):
        delete_model(Todo, todo_id)

    @staticmethod
    @require(lambda todo_id: isinstance(todo_id, int))
    def read(todo_id: int) -> "Todo":
        return get_model(Todo, todo_id)


class Journal(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    uid: str = Field(description="Globally unique identifier")
    dtstamp: datetime = Field(description="Date/time stamp")
    dtstart: Optional[datetime] = Field(
        description="Start date/time of the journal entry"
    )
    summary: Optional[str] = Field(description="Summary of the journal entry")
    description: Optional[str] = Field(description="Description of the journal entry")
    calendar_id: Optional[int] = Field(default=None, foreign_key="calendar.id")
    calendar: Optional[Calendar] = Relationship(back_populates="journals")


class Alarm(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    action: str = Field(description="Action of the alarm")
    trigger: datetime = Field(description="Trigger time of the alarm")
    duration: Optional[str] = Field(description="Duration of the alarm")
    repeat: Optional[int] = Field(description="Repeat count of the alarm")
    description: Optional[str] = Field(description="Description of the alarm")
    event_id: Optional[int] = Field(default=None, foreign_key="event.id")
    event: Optional[Event] = Relationship(back_populates="alarms")

    @staticmethod
    @require(lambda action: isinstance(action, str))
    @require(lambda trigger: isinstance(trigger, datetime))
    def create(
            action: str,
            trigger: str | datetime,
            duration: str = None,
            repeat: int = None,
            description: str = None,
            event_id: Optional[int] = None
    ) -> "Alarm":
        new_alarm = Alarm(
            action=action,
            trigger=parse_datetime(trigger),
            duration=duration,
            repeat=repeat,
            description=description,
            event_id=event_id
        )

        add_model(new_alarm)

        return new_alarm

    @staticmethod
    @require(lambda alarm_id: isinstance(alarm_id, int))
    def update(
            alarm_id: int,
            action: str,
            trigger: str | datetime,
            duration: str = None,
            repeat: int = None,
            description: str = None,
            event_id: Optional[int] = None
    ):
        with update_model(Alarm, alarm_id) as alarm:
            alarm.action = action
            alarm.trigger = parse_datetime(trigger)
            alarm.duration = duration
            alarm.repeat = repeat
            alarm.description = description
            alarm.event_id = event_id

    @staticmethod
    @require(lambda alarm_id: isinstance(alarm_id, int))
    def delete(alarm_id: int):
        delete_model(Alarm, alarm_id)

    @staticmethod
    @require(lambda alarm_id: isinstance(alarm_id, int))
    def read(alarm_id: int) -> "Alarm":
        return get_model(Alarm, alarm_id)



import streamlit as st
from st_pages import add_page_title

from dspygen.modules.streamlit_bot_module import streamlit_bot_call
from dspygen.utils.dspy_tools import init_dspy

add_page_title()


# Streamlit form and display
st.title("StreamlitBotModule Generator")
project = st.text_input("Enter project")
page = st.text_input("Enter page")

if st.button("Submit StreamlitBotModule"):
    init_dspy()

    result = streamlit_bot_call(project=project, page=page)
    st.code(result)


import os
from enum import Enum, auto

import dspy

from dspygen.utils.dspy_tools import init_dspy

from dspygen.mixin.fsm.fsm_mixin import FSMMixin, trigger
from dspygen.modules.condition_sufficient_info_module import condition_sufficient_info_call
from dspygen.modules.query_generator_module import query_generator_call
#from dspygen.modules.refine_results_module import refine_results_module_call
from dspygen.modules.source_selector_module import source_selector_call
from dspygen.utils.scraping_tools import execute_brave_search_queries, scrape_urls, execute_duckduckgo_queries


class LearningAgentState(Enum):
    """ Enum representing the states in the learning objectives lifecycle. """
    INPUT_OBJECTIVES = auto()
    GENERATE_QUERIES = auto()
    EXECUTE_SEARCH = auto()
    SELECT_URLS = auto()
    SCRAPE_AND_CONVERT = auto()
    DISCARD_IRRELEVANT = auto()
    DECIDE_SUFFICIENCY = auto()
    REGENERATE_QUERIES = auto()
    COMPLETING_TASK = auto()
    END = auto()

class LearningAgent(FSMMixin):
    def __init__(self, objectives):
        # super().setup_fsm(LearningAgentState, initial=LearningAgentState.EXECUTE_SEARCH)
        super().setup_fsm(LearningAgentState, initial=LearningAgentState.INPUT_OBJECTIVES)
        self.objectives = objectives
        self.current_objective = None
        self.queries = {}
        self.search_results = {}
        self.selected_urls = {}
        self.scraped_content = {}
        self.refined_content = {}
        self.sufficient_info = {}

    @trigger(source=LearningAgentState.INPUT_OBJECTIVES, dest=LearningAgentState.GENERATE_QUERIES)
    def process_input(self):
        """Processes the initial input of learning objectives."""
        # Placeholder: processing input and setting up initial data
        # print("Processing input objectives.")

    @trigger(source=LearningAgentState.GENERATE_QUERIES, dest=LearningAgentState.EXECUTE_SEARCH)
    def generate_queries(self):
        """Generate search queries for each learning objective using the query_generator module."""
        self.queries = {objective: query_generator_call(objective) for objective in self.objectives}
        # print(f"Generated search queries: {self.queries}")

    @trigger(source=LearningAgentState.EXECUTE_SEARCH, dest=LearningAgentState.SELECT_URLS)
    def execute_search(self):
        """Execute the generated search queries using the Brave Search API."""
        # api_key = os.getenv("BRAVE_API_KEY")
        # self.search_results = execute_brave_search_queries(self.queries, api_key)
        self.search_results = execute_duckduckgo_queries(self.queries)
        print(f"Search results: {self.search_results}")

    @trigger(source=LearningAgentState.SELECT_URLS, dest=LearningAgentState.SCRAPE_AND_CONVERT)
    def select_urls(self):
        """Select URLs based on search results using the source_selector module."""
        # Assuming self.search_results is populated with search results for each objective
        self.selected_urls = {objective: source_selector_call(self.search_results[objective]) for objective in self.objectives}
        print(f"Selected URLs: {self.selected_urls}")

    @trigger(source=LearningAgentState.SCRAPE_AND_CONVERT, dest=LearningAgentState.DISCARD_IRRELEVANT)
    def scrape_and_convert(self):
        """Scrape the selected URLs and convert to text."""
        for objective, urls in self.selected_urls.items():
            # Extract only the URLs to pass to the scrape_urls function
            url_list = [url['url'] for url in urls]
            self.scraped_content[objective] = scrape_urls(url_list)
        print(f"Scraped and converted content.")

    @trigger(source=LearningAgentState.DISCARD_IRRELEVANT, dest=LearningAgentState.DECIDE_SUFFICIENCY)
    def discard_irrelevant(self):
        """Directly relay the scraped content without refining."""
        self.refined_content = self.scraped_content
        print(f"Refined content: {self.refined_content}")
        
        
        #self.refined_content = {objective: refine_results_module_call(self.scraped_content[objective]) for objective in self.objectives}
        #print(f"Refined content: {self.refined_content}")

    @trigger(source=LearningAgentState.DECIDE_SUFFICIENCY, dest=LearningAgentState.COMPLETING_TASK, conditions=['is_information_sufficient'])
    def decide_sufficiency(self):
        """Decide if the information gathered for each learning objective is sufficient."""
        for objective, contents in self.refined_content.items():
            # Assuming `contents` is a list of contents from different sources
            # Evaluate the combined sufficiency across all sources for the current objective
            sufficiency_results = [condition_sufficient_info_call(objective, content) for content in contents]
            # Objective is considered sufficient if all source checks return 1
            self.sufficient_info[objective] = all(sufficiency_results)
            print(f"Information for '{objective}' sufficient across all sources: {self.sufficient_info[objective]}")

    @trigger(source=LearningAgentState.DECIDE_SUFFICIENCY, dest=LearningAgentState.REGENERATE_QUERIES, unless=['is_information_sufficient'])
    def regenerate_queries(self):
        """Regenerate queries for insufficient information and limit to one loop."""
        # Placeholder: regenerating queries
        print("Regenerating queries for insufficient information.")

    @trigger(source=LearningAgentState.REGENERATE_QUERIES, dest=LearningAgentState.EXECUTE_SEARCH)
    def loop_back(self):
        """Loop back to execute search with new queries."""
        print("Looping back to execute search with new queries.")

    @trigger(source=LearningAgentState.COMPLETING_TASK, dest=LearningAgentState.END)
    def complete_task(self):
        """Complete the learning task."""
        print("Task completed.")

    def is_information_sufficient(self):
        """Check if the information gathered is sufficient across all objectives."""
        # Check if all values in sufficient_info are True
        return all(self.sufficient_info.values())


def main():
    # Example objectives
    init_dspy(dspy.OllamaLocal, model="llama3")
    objectives = ["how to finetune a llama 3 lora model"]
    agent = LearningAgent(objectives)
    print("Initial state:", agent.state)

    # Manually triggering each step to simulate workflow
    agent.process_input()
    agent.generate_queries()
    agent.execute_search()
    agent.select_urls()
    agent.scrape_and_convert()
    agent.discard_irrelevant()
    agent.decide_sufficiency()

    # Check if there is a need to regenerate queries
    if not agent.is_information_sufficient():
        agent.regenerate_queries()
        agent.execute_search()
        agent.select_urls()
        agent.scrape_and_convert()
        agent.discard_irrelevant()
        agent.decide_sufficiency()

    agent.complete_task()
    print("Final state:", agent.state)

if __name__ == "__main__":
    main()




"""dspygen CLI."""
import json
import sys
from importlib import import_module, metadata
import subprocess
import os

from pathlib import Path

import typer
from munch import Munch

from dspygen.utils.cli_tools import chatbot
from dspygen.utils.file_tools import source_dir
from dspygen.utils.module_tools import module_to_dict

app = typer.Typer()


# Load existing subcommands
def load_commands(directory: str = "subcommands"):
    script_dir = Path(__file__).parent
    subcommands_dir = script_dir / directory

    for filename in os.listdir(subcommands_dir):
        if filename.endswith("_cmd.py"):
            module_name = f'{__name__.split(".")[0]}.{directory}.{filename[:-3]}'
            module = import_module(module_name)
            if hasattr(module, "app"):
                app.add_typer(module.app, name=filename[:-7])


def package_installed(package_name, min_version):
    try:
        version = metadata.version(package_name)
        return version >= min_version
    except metadata.PackageNotFoundError:
        return False


def check_or_install_packages():
    packages_requirements = {
        "cruft": "2.12.0",
        "cookiecutter": "2.1.1",
    }

    for package, min_version in packages_requirements.items():
        if not package_installed(package, min_version):
            print(f"{package} not found or version is below {min_version}. Installing/upgrading...")
            subprocess.check_call([sys.executable, "-m", "pip", "install", f"{package}>={min_version}"])
        # else:
        #     print(f"{package} meets the version requirement.")


@app.command()
def init(project_name: str = typer.Argument(...),
         author_email: str = typer.Argument("todo@todo.com"),
         author_name: str = typer.Argument("TODO")):
    """Initialize the DSPygen project."""
    # If the project has underscores or spaces throw an error
    if "_" in project_name or " " in project_name:
        print("Project name should not contain underscores or spaces.")
        sys.exit(1)
    elif project_name[0] == "-" or project_name[0] == "_":
        print("Project name should not start with a hyphen or underscore.")
        sys.exit(1)

    check_or_install_packages()

    extra_context = Munch(project_name=project_name,
                          author_email=author_email,
                          author_name=author_name)


    # The template URL and the configuration for the new project
    template_url = "https://github.com/radix-ai/poetry-cookiecutter"
    # Project initialization logic, assuming static configuration for demonstration
    try:
        print(f"Creating new project named {project_name}...")
        subprocess.check_call(["cruft", "create", template_url,
                               "--config-file", source_dir("config.yaml"),
                               "--extra-context", f'{json.dumps(extra_context)}',
                               "--no-input"])

        # We need to install dspygen in the project's virtual environment
        # It uses poetry to manage the virtual environment
        # Change to the project directory
        # Run the command to initialize the virtual environment
        # Run the command to install dspygen in the virtual environment

        os.chdir(project_name)

        subprocess.check_call(["poetry", "install"])
        # Create the virtual environment
        subprocess.check_call(["poetry", "env", "use", "python"])
        # Install the project in the virtual environment
        subprocess.check_call(["poetry", "add", "dspygen"])

        subprocess.check_call(["poetry", "run", "pip", "install", "-e", "."])
        # Change back to the original directory
        os.chdir("..")

        print(f"Project {project_name} initialized successfully.")
    except subprocess.CalledProcessError:
        print("Failed to initialize the new project.")
        sys.exit(1)


TUTOR_CONTEXT = """DSPyGen: AI Development Simplified
DSPyGen revolutionizes AI development by bringing the "Convention over Configuration" philosophy to language model (LM) pipelines. Inspired by Ruby on Rails, it offers a CLI for creating, developing, and deploying with DSPy modules, emphasizing quick setup and modular design for streamlined project workflows.

Key Features:

Quick Initialization: Rapidly configure your AI project, mirroring the simplicity of Ruby on Rails.
Modular Design: Generate and enhance DSPy modules with ease, promoting a scalable and flexible development environment.
User-Friendly Commands: Manage your AI projects effortlessly through an intuitive command structure.
Chatbot Assistance: Embedded support to guide through the development process, enhancing user experience.
Using DSPyGen Modules:
DSPyGen's core lies in its modules, designed for seamless integration and code optimization. Here’s how to leverage them:

dspygen is a command-line tool.
It helps generate various components for a project.
Usage includes options and commands.
Options like --install-completion, --show-completion, and --help are available.
Commands include actor, assert, browser, command, help, init, lm, module, sig, and tutor.
Each command serves a specific purpose:
actor: Related to actors.
assert: Generates assertions for dspy.
browser: Pertains to browser functionality.
command: Generates or adds subcommands.
help: Provides assistance and updates help files.
init: Initializes a DSPygen project.
lm: Generates language models.
module: Deals with generating or calling DSPy modules.
sig: Generates dspy.Signatures.
tutor: Guides through project development with DSPyGen.

blog: Calls modules related to blogging.
book_appointment: Invokes modules for scheduling appointments.
chat_bot: Initiates module calls for chatbots.
checker: Calls modules for checking or validating.
choose_function: Initiates module calls involving choice or selection functions.
dflss: Related to Six Sigma methodology, possibly for invoking process improvement modules.
gen_cli: Triggers modules for command-line interfaces.
gen_dspy: Calls modules specific to DSPy.
gen_keyword_arguments: Initiates module calls with keyword arguments.
gen_signature: Invokes signature-related modules.
html: For modules related to HTML.
insight_tweet: Calls modules for generating insightful tweets.
message: Initiates module calls for messaging.
module_docstring: Triggers module calls for documenting modules.
product_bot: Invokes module calls for product-related bots.
prompt_function_call: Triggers modules for prompting function calls.
python_expert: For modules related to Python expertise.
python_source_code: Initiates module calls for generating Python source code.
source_code_pep8_docs: Triggers module calls for source code with PEP8 documentation.
subject_destination_audience_newsletter_article: Initiates module calls for newsletter articles.
text_summary_module: Calls modules for summarizing text.
to_elixir: Initiates module calls to convert modules to Elixir format.

Use this information to guide the usage of the DSPyGen CLI and its modules.
"""

@app.command(name="tutor")
def tutor(question: str = ""):
    """Guide you through developing a project with DSPyGen."""
    from dspygen.utils.dspy_tools import init_dspy
    init_dspy(max_tokens=3000, model="gpt-4")

    chatbot(question, TUTOR_CONTEXT)


def main():
    print("Welcome to DSPyGen CLI!")
    # init("test_project")

    # import json
    # import yaml

    # current_module = sys.modules[__name__]
    # module_dict = module_to_dict(current_module, include_docstring=False)
    # print(yaml.dump(module_dict))
    # print(json.dumps(module_dict, indent=2))



if __name__ == '__main__':
    main()
else:
    load_commands()



import os

from dspygen.utils.file_tools import pages_dir


def main():
    print(pages_dir())
    # print(pages_dir("hello.py"))

    print(os.path.exists(pages_dir()))
    # print(os.path.exists(pages_dir("hello.py"))


if __name__ == '__main__':
    main()



import json
from pathlib import Path
from typing import TypeVar, Generic, Type, List, Optional

from pydantic import BaseModel

T = TypeVar('T', bound=BaseModel)


class BaseRepository(Generic[T]):
    def __init__(self, model: Type[T], storage_file: Path):
        self.model = model
        self.storage_file = storage_file
        self.storage_file.touch(exist_ok=True)  # Ensure file exists

    def _read_data(self) -> List[T]:
        """Reads and returns all model instances from the storage file."""
        try:
            with self.storage_file.open('r', encoding='utf-8') as file:
                data = json.load(file)
            return [self.model.parse_obj(item) for item in data]
        except (json.JSONDecodeError, FileNotFoundError):
            return []

    def _write_data(self, data: List[T]) -> None:
        """Writes the provided list of model instances to the storage file."""
        with self.storage_file.open('w', encoding='utf-8') as file:
            json.dump([item.dict() for item in data], file, indent=4)

    def add(self, item: T) -> None:
        """Adds a new model instance to the storage."""
        items = self._read_data()
        items.append(item)
        self._write_data(items)

    def get(self, **criteria) -> Optional[T]:
        """Retrieves a single model instance that matches the provided criteria."""
        items = self._read_data()
        for item in items:
            if all(getattr(item, field) == value for field, value in criteria.items()):
                return item
        return None

    def remove(self, **criteria) -> bool:
        """Removes model instances that match the provided criteria."""
        items = self._read_data()
        items_before = len(items)
        items = [item for item in items if not all(getattr(item, field) == value for field, value in criteria.items())]
        self._write_data(items)
        return len(items) < items_before

    def update(self, item: T) -> None:
        """Updates a model instance in the storage."""
        items = self._read_data()
        for i, existing_item in enumerate(items):
            if existing_item.id == item.id:  # Assuming each model has a unique 'id' field
                items[i] = item
                break
        self._write_data(items)

    def save(self, item: T) -> None:
        """Updates a model instance if it exists, otherwise inserts it."""
        items = self._read_data()
        for i, existing_item in enumerate(items):
            if existing_item.id == item.id:  # Assuming each model has a unique 'id' field
                items[i] = item
                self._write_data(items)
                return
        self.add(item)

    def list_all(self) -> List[T]:
        """Returns a list of all model instances."""
        return self._read_data()


def main():
    """Main function"""
    # Conversation Learnng Through Dialog Repository

    class ConversationRepository(BaseRepository[Conversation]):
        def __init__(self, storage_file: Path):
            super().__init__(Conversation, storage_file)


if __name__ == '__main__':
    main()






# Signature for generating mocks based on classes and methods
import dspy
from dspy.teleprompt import MIPRO


class GenerateMocks(dspy.Signature):
    """Generate mocks for classes and functions."""
    class_names = dspy.InputField(desc="List of classes to generate mocks for.")
    method_names = dspy.InputField(desc="List of methods to generate mocks for.")
    mocks = dspy.OutputField(desc="Generated mocks for the input classes and methods.")


class GenerateTestCases(dspy.Signature):
    """Generate test cases for the given mocks."""
    mocks = dspy.InputField(desc="The mocks generated for the buggy file.")
    test_cases = dspy.OutputField(desc="Generated test cases.")


# Define a module to create and optimize the mocks and tests
class AutoPytestMockRover(dspy.Module):
    def __init__(self):
        super().__init__()
        self.generate_mocks = dspy.ChainOfThought("class_names,method_names->mocks")
        self.generate_test_cases = dspy.ChainOfThought("mocks->test_cases")

    def forward(self, class_names, method_names):
        mocks = self.generate_mocks(class_names=str(class_names), method_names=str(method_names)).mocks
        test_cases = self.generate_test_cases(mocks=mocks).test_cases
        return dspy.Prediction(mocks=mocks, test_cases=test_cases)


def main():
    """Main function"""
    from dspygen.utils.dspy_tools import init_ol
    init_ol()

    apmr = AutoPytestMockRover()
    class_names = ["ChatGPTRetriever", "ChatGPTChromaDBRetriever"]
    method_names = ["forward", "prepare_queries"]
    result = apmr.forward(class_names=class_names, method_names=method_names)
    print(result)


if __name__ == '__main__':
    main()



import dspy
from dspygen.modules.file_name_module import file_name_call


class CodeWriter():
    def __init__(self, **kwargs):
        super().__init__()
    
    def forward(self, **kwargs):
        return None


def code_writer_call(source):
    rm = CodeWriter()
    return rm.forward(source=source)


def main():
    rm = CodeWriter()
    print(rm.forward())
    
    
if __name__ == '__main__':
    main()
      


from pathlib import Path

import typer

from dspygen.typetemp.functional import render

app = typer.Typer(help="""Generate new sub commands or add to existing ones.""")


subcommand_template = '''"""{{ subcommand_name }}"""
import typer


app = typer.Typer()


@app.command(name="{{ new_command_name }}")
def {{ sub_command_name }}_{{ new_command_name }}():
    """{{ new_command_name }}"""
    typer.echo("Running {{ new_command_name }} subcommand.")
    
'''


# Define the subcommand to generate subcommand modules
@app.command(
    name="new",
)
def module(subcommand_name: str, new_command_name: str):
    """
    Generate a new subcommand module with the given name.
    Example usage: dspygen command new new_command
    """
    script_dir = Path(__file__).parent

    # Generate the filename for the new subcommand module
    filename = f"{subcommand_name}_cmd.py"
    module_path = script_dir / filename

    # Check if the existing subcommand module file exists
    if module_path.exists():
        typer.echo(f"Subcommand module '{subcommand_name}' already exists.")
        return

    # Create the subcommand module file
    with open(script_dir / filename, "w") as file:
        # You can customize the content of the module here
        source = render(subcommand_template, subcommand_name=subcommand_name, new_command_name=new_command_name)
        file.write(source)

    typer.echo(f"Subcommand module '{subcommand_name}' generated successfully!")


add_template = ''' 
@app.command(name="{{ new_command_name }}")
def {{ sub_command_name }}_{{ new_command_name }}():
    """{{ new_command_name }}"""
    typer.echo("Running {{ new_command_name }} subcommand.")

'''


@app.command(name="add")
def add_command(sub_command_name: str, new_command_name: str):
    """
    Add a new command to an existing subcommand module.
    Example usage: dspygen command add existing_command new_command
    """
    script_dir = Path(__file__).parent

    # Construct the filename for the existing subcommand module
    filename = f"{sub_command_name}_cmd.py"

    # Construct the path to the existing subcommand module
    module_path = script_dir / filename

    # Check if the existing subcommand module file exists
    if not module_path.exists():
        typer.echo(f"Subcommand module '{sub_command_name}' does not exist.")
        return

    # Append the code to the existing subcommand module file
    with open(module_path, "a") as module_file:
        new_command_code = render(
            add_template,
            sub_command_name=sub_command_name,
            new_command_name=new_command_name,
        )
        module_file.write(new_command_code)

    typer.echo(
        f"New command '{new_command_name}' added to subcommand module '{sub_command_name}' successfully!"
    )



import typer
from transitions import Machine
from typing import List
import json

app = typer.Typer()

# Value Objects
class Topic:
    def __init__(self, name: str, content: str):
        self.name = name
        self.content = content

class Question:
    def __init__(self, question: str, options: List[str], answer: str):
        self.question = question
        self.options = options
        self.answer = answer

# Domain Entities
class TutoringSession:
    def __init__(self, topics: List[Topic], questions: List[Question], current_topic_index: int = 0, score: int = 0):
        self.topics = topics
        self.questions = questions
        self.current_topic_index = current_topic_index
        self.score = score
        self.machine = Machine(model=self, states=['introduction', 'topic_exploration', 'quiz', 'completed'], initial='introduction')
        self.machine.add_transition('start_topic', 'introduction', 'topic_exploration')
        self.machine.add_transition('next_topic', '*', None, after='move_to_next_topic')
        self.machine.add_transition('start_quiz', 'topic_exploration', 'quiz', unless=['has_next_topic'])
        self.machine.add_transition('complete', 'quiz', 'completed')

    def has_next_topic(self):
        return self.current_topic_index < len(self.topics) - 1

    def get_current_topic(self) -> Topic:
        return self.topics[self.current_topic_index]

    def move_to_next_topic(self):
        if self.has_next_topic():
            self.current_topic_index += 1

    def explore_topic(self):
        current_topic = self.get_current_topic()
        typer.echo(f"Topic: {current_topic.name}")
        typer.echo(current_topic.content)

    def start_quiz(self):
        typer.echo("Let's put your knowledge to the test with a quiz!")
        typer.echo("Answer the following questions based on what you've learned.")

        for question in self.questions:
            typer.echo(question.question)
            for option in question.options:
                typer.echo(option)
            user_answer = typer.prompt("Enter your answer (A/B/C/D):")
            if user_answer.upper() == question.answer:
                typer.echo("Correct!")
                self.score += 1
            else:
                typer.echo(f"Incorrect. The correct answer is {question.answer}.")

        typer.echo(f"Quiz completed. Your score: {self.score}/{len(self.questions)}")

# Root Aggregate
class TutoringSessionAggregate:
    def __init__(self, session: TutoringSession):
        self.session = session

    def start_session(self):
        self.session.start_topic()

    def next_topic(self):
        if self.session.has_next_topic():
            self.session.move_to_next_topic()
        else:
            self.session.start_quiz()

    def complete_session(self):
        self.session.complete()

# Repository
class TutoringSessionRepository:
    def __init__(self, file_path: str):
        self.file_path = file_path

    def load_session(self) -> TutoringSessionAggregate:
        try:
            with open(self.file_path, 'r') as file:
                data = json.load(file)
                topics = [Topic(topic['name'], topic['content']) for topic in data['topics']]
                questions = [Question(question['question'], question['options'], question['answer']) for question in data['questions']]
                session = TutoringSession(topics, questions, data['current_topic_index'], data['score'])
                return TutoringSessionAggregate(session)
        except FileNotFoundError:
            typer.echo("No existing session found. Starting a new session.")
            return self.create_new_session()

    def save_session(self, session_aggregate: TutoringSessionAggregate):
        session = session_aggregate.session
        data = {
            'topics': [{'name': topic.name, 'content': topic.content} for topic in session.topics],
            'questions': [{'question': question.question, 'options': question.options, 'answer': question.answer} for question in session.questions],
            'current_topic_index': session.current_topic_index,
            'score': session.score
        }
        with open(self.file_path, 'w') as file:
            json.dump(data, file)

    def create_new_session(self) -> TutoringSessionAggregate:
        topics = [
            Topic("Topic 1: Introduction to Philosophy", "Philosophy is the study of fundamental questions about existence, knowledge, values, reason, and the mind."),
            Topic("Topic 2: Socratic Method", "The Socratic method is a form of dialogue and questioning used to stimulate critical thinking and draw out ideas.")
        ]
        questions = [
            Question("What is philosophy?", [
                "A. The study of fundamental questions about existence, knowledge, values, reason, and the mind",
                "B. A branch of science that deals with the natural world",
                "C. The practice of making persuasive arguments",
                "D. The art of creating beautiful works"
            ], "A"),
            Question("What is the goal of the Socratic method?", [
                "A. To memorize facts and information",
                "B. To engage in small talk and casual conversation",
                "C. To deepen understanding, encourage self-reflection, and facilitate the discovery of truth",
                "D. To prove one's superiority over others"
            ], "C")
        ]
        session = TutoringSession(topics, questions)
        return TutoringSessionAggregate(session)

@app.command()
def tutor(file_path: str = "session_data.json"):
    """
    Processes commands to navigate through the tutoring session and quiz.
    """
    repository = TutoringSessionRepository(file_path)
    session_aggregate = repository.load_session()
    session = session_aggregate.session

    while True:
        if session.state == 'introduction':
            typer.echo("Welcome to the Socratic Method Text Adventure!")
            typer.echo("In this game, you will explore philosophical concepts and engage in Socratic dialogue.")
            command = typer.prompt("Enter /start to begin:")
            if command == "/start":
                session_aggregate.start_session()
        elif session.state == 'topic_exploration':
            topic = session.get_current_topic()
            typer.echo(f"Current Topic: {topic.name}")
            session.explore_topic()
            command = typer.prompt("Enter /next to move to the next topic, or /quiz to start the quiz:")
            if command == "/next":
                session_aggregate.next_topic()
                if not session.has_next_topic():
                    session.start_quiz()
            elif command == "/quiz":
                session.start_quiz()
            else:
                typer.echo("Unknown command. Please try again.")
        elif session.state == 'quiz':
            session.start_quiz()
            command = typer.prompt("Enter /complete to finish the tutoring session:")
            if command == "/complete":
                session_aggregate.complete_session()
                repository.save_session(session_aggregate)
                typer.echo("Tutoring session completed. Thank you for participating!")
                break
            else:
                typer.echo("Unknown command. Please try again.")
        elif session.state == 'completed':
            break
        else:
            typer.echo("Unknown state. Please try again.")

        repository.save_session(session_aggregate)

if __name__ == "__main__":
    app()


import dspy
import typing

from pydantic import BaseModel, Field


Model = typing.TypeVar('Model', bound='BaseModel')


class GenerateJSONForPrompt(dspy.Signature):
    """
    Takes a prompt and a response schema and returns a JSON formatted prompt. Make sure the schema is completely
    aligned with the prompt and schema
    """
    prompt = dspy.InputField(desc="Text prompt for generating the response.")
    response_schema = dspy.InputField(desc="Schema describing the desired structure of the response.")

    json_for_prompt = dspy.OutputField(desc="JSON formatted prompt based on input specifications.",
                                       prefix="Only return the JSON formatted prompt based on the input specifications "
                                              "and schema.\n\n```json\n",)


class BaseModelModule(dspy.Module):
    """JsonModule"""
    def __init__(self, **forward_args):
        super().__init__()
        self.forward_args = forward_args
        self.output = None

    def forward(self, model: type[Model], prompt) -> Model:
        pred = dspy.ChainOfThought(GenerateJSONForPrompt)
        from inspect import getsource
        result = pred(response_schema=str(model.model_json_schema()), prompt=prompt).json_for_prompt
        print(result)
        from dspygen.utils.json_tools import extract
        result = extract(result)
        self.output = model.model_validate(result)
        return self.output


# RFC 5545 VEvent
class VEvent(BaseModel):
    dtstart: str
    dtend: str
    summary: str
    location: str
    description: str


class HelloWorldModel(BaseModel):
    guid: str = Field(..., description="The GUID of the message.")
    message: str = Field(..., description="The message to be displayed.")
    world_description: str = Field(..., description="The description of the world. Required")


def model_call(model: type[Model], prompt: str) -> Model:
    """Takes the BaseModel type and prompt and returns the BaseModel instance."""
    model_module = BaseModelModule()
    return model_module.forward(model=model, prompt=prompt)


def main():
    """Main function"""
    from dspygen.utils.dspy_tools import init_ol
    # init_ol(model="llama3")
    init_ol(model="phi3:instruct")
    # pred = dspy.Predict(GenerateJSONForPrompt)
    # result = pred(response_schema=str(HelloWorldModel.model_json_schema()), prompt="Hello, World!").json_for_prompt
    # from dspygen.utils.json_tools import extract
    # result = extract(result)
    # print(result)

    vevent_meeting_desc = """Hello Robin, I would like to schedule a meeting with you.
    The meeting will be held at the conference room on the 5th floor.
    The meeting will start at 10:00 AM and end at 11:00 AM, May 9th 2024.
    The purpose of the meeting is to discuss the upcoming project.
    """

    vevent = model_call(VEvent, vevent_meeting_desc)

    print(vevent)


if __name__ == '__main__':
    main()



import os

import inflection
from slugify import slugify

from dspygen.dsl.dsl_pipeline_executor import execute_pipeline


def feature_code_generation():

    context = execute_pipeline(f'{os.getcwd()}/data_gherkin_pipeline.yaml',
                               init_ctx={"file_path": f"{os.getcwd()}/features.csv"})

    file_name = "hello-world"  # slugify(f"{inflection.underscore(result['FeatureDescription'])}")

    with open(f"{file_name}.tsx", 'w') as f:
        code = context.react_code
        # remove trailing ``` if present
        if code.endswith("```"):
            code = code[:-3]
        f.write(context.react_code)
        print(f"React JSX code written to {file_name}")


def main():
    feature_code_generation()


if __name__ == '__main__':
    main()



"""

"""
import dspy
from dspygen.utils.dspy_tools import init_dspy        


class GenMessageModule(dspy.Module):
    """GenMessageModule"""

    def forward(self, prompt, message):
        pred = dspy.Predict("prompt, message -> pydantic_model_validate_str")
        result = pred(prompt=prompt, message=message).pydantic_model_validate_str
        return result


from typer import Typer
app = Typer()


@app.command()
def call(prompt, message):
    """GenMessageModule"""
    init_dspy()

    print(gen_message_call(prompt=prompt, message=message))



def gen_message_call(prompt, message):
    gen_message = GenMessageModule()
    return gen_message.forward(prompt=prompt, message=message)



def main():
    init_dspy()
    prompt = ""
    message = ""
    print(gen_message_call(prompt=prompt, message=message))



from fastapi import APIRouter
router = APIRouter()

@router.post("/gen_message/")
async def gen_message_route(data: dict):
    # Your code generation logic here
    init_dspy()

    print(data)
    return gen_message_call(**data)



if __name__ == "__main__":
    main()



import dspy
from typer import Typer


app = Typer(help="Create React JSX source code.")


class PromptReactJsxModule(dspy.Module):

    """This is a DSPy Module that converts a prompt into react_jsx"""

    def forward(self, prompt):
        pred = dspy.Predict("prompt -> react_jsx")
        result = pred(prompt=prompt).react_jsx
        return result


def main():
    lm = dspy.OpenAI(max_tokens=500)
    dspy.settings.configure(lm=lm)

    prompt = "Hello World Functional Component"

    prompt_react_jsx = PromptReactJsxModule()
    print(prompt_react_jsx.forward(prompt=prompt))


@app.command()
def module_test(prompt):
    """This is a DSPy Module that converts a prompt into react_jsx"""
    prompt_react_jsx = PromptReactJsxModule()

    print(prompt_react_jsx.forward(prompt=prompt))


if __name__ == "__main__":
    # app()
    main()



from dspygen.dsl.dsl_pipeline_executor import execute_pipeline
from dspygen.dsl.dsl_pydantic_models import *


def main():
    from dspygen.utils.dspy_tools import init_dspy
    init_dspy()

    signature = GenSignatureModel.to_inst("Product Requirements Document for Websites")

    signature.inputs = [signature.inputs[0]]

    module = GenLMModuleModel.to_inst(f"{signature}")
    module.signature = signature.name

    arg_dict = {signature.inputs[0].name: "{{ user_input }}"}

    step = StepDSLModel(module=module.name, signature=signature.name, args=arg_dict)

    pipeline = PipelineDSLModel(signatures=[signature], modules=[module], steps=[step])

    pipeline.to_yaml(f"{signature.name}_pipeline.yaml")

    context = execute_pipeline(f"{signature.name}_pipeline.yaml",
                               {"user_input": "A full stack nextjs DSL pipeline nocode generator"})

    print(context)


if __name__ == '__main__':
    main()



from dspygen.modules.gen_keyword_arguments_module import invoke
from dspygen.utils.dspy_tools import init_dspy


def main():
    chat("bot: Hello world")


def chat(message: str):
    response = chat(invoke(chat, f"{message}\nbot:"))
    print(response)
    return response


if __name__ == '__main__':
    init_dspy()

    main()



from dspygen.mixin.fsm.fsm_mixin import FSMMixin, trigger


from enum import Enum, auto


class CoderAgentState(Enum):
    """ Enum representing the states in the coding task lifecycle. """
    ANALYZING_REQUIREMENTS = auto()
    WRITING_CODE = auto()
    TESTING_CODE = auto()
    REFACTORING_CODE = auto()
    HANDLING_ERRORS = auto()
    COMPLETING_TASK = auto()


class CoderAgent(FSMMixin):
    def __init__(self, initial=CoderAgentState.ANALYZING_REQUIREMENTS):
        super().setup_fsm(CoderAgentState, initial=initial)
        self.git_repo = None
        self.code_base = None
        self.tests = None
        self.errors = None
        self.console_output = None
        # Initialize any state-specific attributes if necessary

    @trigger(
        source=CoderAgentState.ANALYZING_REQUIREMENTS,
        dest=CoderAgentState.WRITING_CODE,
        # conditions=['requirements_met'],
        after=['notify_progress']
    )
    def start_coding(self):
        """Transition from analyzing requirements to writing code."""
        print("Starting to write code.")

    @trigger(
        source=CoderAgentState.WRITING_CODE,
        dest=CoderAgentState.TESTING_CODE,
        conditions=['code_complete'],
        after=['run_tests']
    )
    def test_code(self):
        """Transition from writing code to testing code."""
        print("Testing code now.")

    @trigger(
        source=CoderAgentState.TESTING_CODE,
        dest=CoderAgentState.HANDLING_ERRORS,
        conditions=['errors_detected'],
        unless=['all_tests_passed']
    )
    def handle_errors(self):
        """Handle errors if any are detected during testing."""
        print("Handling coding errors.")

    @trigger(
        source=[CoderAgentState.HANDLING_ERRORS, CoderAgentState.TESTING_CODE],
        dest=CoderAgentState.REFACTORING_CODE,
        conditions=['errors_resolved'],
        after=['optimize_code']
    )
    def refactor_code(self):
        """Refactor code after errors are resolved."""
        print("Refactoring code.")

    @trigger(
        source=[CoderAgentState.TESTING_CODE, CoderAgentState.REFACTORING_CODE],
        dest=CoderAgentState.COMPLETING_TASK,
        # conditions=['optimization_complete'],
        after=['finalize_task']
    )
    def complete_task(self):
        """Complete the coding task after successful refactoring."""
        print("Task completed.")

    # Placeholder methods for conditions and actions
    def requirements_met(self):
        return True

    def notify_progress(self):
        print("Notifying progress on the task.")

    def code_complete(self):
        return True

    def run_tests(self):
        print("Running tests on the code.")

    def errors_detected(self):
        return False

    def all_tests_passed(self):
        return True

    def errors_resolved(self):
        return True

    def optimize_code(self):
        print("Optimizing the code.")

    def optimization_complete(self):
        return True

    def finalize_task(self):
        print("Finalizing and wrapping up the task.")


def main():
    # Create an instance of the CoderAgent starting at the initial state
    coder = CoderAgent()

    # Print the initial state
    print(f"Initial State: {coder.state}")

    # Start coding by transitioning from analyzing requirements to writing code
    coder.start_coding()
    print(f"State after starting to code: {coder.state}")

    # Test the code by transitioning from writing code to testing code
    coder.test_code()
    print(f"State after testing code: {coder.state}")

    # Assuming tests pass, move directly to completing task
    # Normally, you would handle errors if they occurred, but this is the happy path
    coder.complete_task()
    print(f"State after completing the task: {coder.state}")

    # Finalize and print confirmation of task completion
    print("Task has been successfully completed!")

# This is a protective measure to ensure the main function is executed only when the script is run directly,
# not when imported as a module.
if __name__ == "__main__":
    main()



from dspygen.typetemp.template.render_mixin import RenderMixin


class TypedTemplate(RenderMixin):
    """Base class for creating templated classes. Uses the jinja2 templating engine
    to render templates. Allows for usage of macros and filters.
    """

    template_path: str = ""  # The path to the file
    source: str = ""  # The string template to be rendered
    to: str = ""  # The "to" property for rendering destination
    output: str = ""  # The rendered output

    def __init__(self, **kwargs):
        if kwargs.get("template_path"):
            # Read the template to the source string
            with open(kwargs.get("template_path")) as f:
                self.source = f.read()

        self.__dict__.update(kwargs)

    def __call__(self, use_native=False, **kwargs) -> str:
        # Use NativeEnvironment when use_native is True, else use default Environment
        return self._render(use_native, **kwargs)

    def render(self, use_native=False, **kwargs) -> str:
        return self._render(use_native, **kwargs)



import os
import tempfile
from typing import Optional

from fastapi import APIRouter, HTTPException
from pydantic import BaseModel
from sqlalchemy.dialects.mysql import json

from dspygen.workflow.workflow_executor import execute_workflow
from dspygen.workflow.workflow_models import Workflow

router = APIRouter()


class WorkflowRequest(BaseModel):
    yaml_content: str
    init_ctx: Optional[dict] = None


# Route to execute a workflow based on YAML content and an optional initial context.
@router.post("/execute_workflow/")
async def run_workflow(request: WorkflowRequest):
    try:
        # Create a temporary file to store the YAML content for processing.
        with tempfile.NamedTemporaryFile(delete=False, mode="w+", suffix=".yaml") as tmp:
            tmp.write(request.yaml_content)
            tmp_path = tmp.name

        # Load the workflow from the temporary YAML file.
        wf = Workflow.from_yaml(tmp_path)

        # Execute the workflow with the provided initial context.
        context = execute_workflow(wf, request.init_ctx)

        # Clean up the temporary file after execution.
        os.remove(tmp_path)

        # Ensure the context returned is JSON serializable by filtering non-serializable objects.
        serializable_context = {k: v for k, v in context.items() if json_serializable(v)}

        return serializable_context
    except Exception as e:
        if "tmp_path" in locals():
            os.remove(tmp_path)
        raise HTTPException(status_code=500, detail=str(e))


def json_serializable(value):
    """Check if value is JSON serializable."""
    try:
        json.dumps(value)
        return True
    except TypeError:
        return False


import dspy
from pydantic import BaseModel, Field
from typer import Typer

from dspygen.typetemp.functional import render

app = Typer()


class DSPyModuleTemplate(BaseModel):
    """{{ input or list of inputs }} -> {{ outputs }}"""

    inputs: list[str] = Field(..., description="Inputs for dspy.Module jinja template.")
    output: str = Field(..., description="Output for dspy.Module jinja template.")
    class_name: str = Field(
        ..., description="Class name combining the inputs and outputs"
    )


dspy_module_template = '''"""
{{ module_name }}
"""
import dspy
from dspygen.utils.dspy_tools import init_dspy

{%- set module_name = model.class_name | class_name ~ "Module" %}
{%- set inputs_join = model.inputs | join(', ') %}    
{%- set inputs_join_kwargs = model.inputs | map('to_kwarg') | join(', ') %}        
{% set var_name = model.class_name | var_name %}


class {{ module_name }}(dspy.Module):
    """{{ module_name }}"""
    
    def __init__(self, **forward_args):
        super().__init__()
        self.forward_args = forward_args
        self.output = None
        
    def __or__(self, other):
        if other.output is None and self.output is None:
            self.forward(**self.forward_args)

        other.pipe(self.output)

        return other

    def forward(self, {{ inputs_join }}):
        pred = dspy.Predict("{{ inputs_join }} -> {{ model.output }}")
        self.output = pred({{ inputs_join_kwargs }}).{{ model.output }}
        return self.output
        
    def pipe(self, input_str):
        raise NotImplementedError("Please implement the pipe method for DSL support.")
        # Replace TODO with a keyword from you forward method
        # return self.forward(TODO=input_str)


{% include 'dspy_module_cli_call.j2' %}



{% include 'dspy_module_def_call.j2' %}



{% include 'dspy_module_main.j2' %}



{% include 'dspy_module_route.j2' %}



{% include 'dspy_module_streamlit_input.j2' %}


if __name__ == "__main__":
    main()

'''


class SignatureDspyModuleModule(dspy.Module):
    """SignatureDspyModuleModule"""

    def forward(self, tmpl_model):
        source = render(dspy_module_template, model=tmpl_model, docstring="")

        return source




from typing import Any

import jinja2

from ..environment.typed_environment import async_environment, environment
from ..environment.typed_native_environment import (
    async_native_environment,
    native_environment,
)


def render_str(source, **kwargs) -> str:
    try:
        template = environment.from_string(source)
    except jinja2.exceptions.TemplateSyntaxError as e:
        print(f"Error processing template: {e}")
        print(f"Problematic template string: {source}")
        raise
    return template.render(**kwargs)


async def arender_str(source, **kwargs) -> str:
    template = async_environment.from_string(source)

    return await template.render_async(**kwargs)


def render_py(source, env=native_environment, **kwargs) -> Any:
    template = env.from_string(source)

    return template.render(**kwargs)


async def arender_py(source, env=async_native_environment, **kwargs) -> Any:
    template = env.from_string(source)

    return await template.render_async(**kwargs)
